import datetime as dt

urls = {"call":
        {"dl": {
            "main": "https://web.bevsa.com.uy/Mercado/MercadoDinero/CallHistoricoDiario.aspx"},
            "source": {"direct": [],
                       "indirect": [
                       "https://web.bevsa.com.uy/Mercado/MercadoDinero/CallHistoricoDiario.aspx"],
                       "provider": ["BEVSA"]}},
        "bonds":
        {"dl": {
            "usd": "https://web.bevsa.com.uy/CurvasVectorPrecios/CurvasIndices/Indices/IndiceITBGL.aspx",
            "ui": "https://web.bevsa.com.uy/CurvasVectorPrecios/CurvasIndices/Indices/IndiceINDUI.aspx",
            "uyu": "https://web.bevsa.com.uy/CurvasVectorPrecios/CurvasIndices/Indices/IndiceITLUP.aspx"},
            "source": {"direct": [],
                       "indirect": [
                           "https://web.bevsa.com.uy/CurvasVectorPrecios/CurvasIndices/Indices/IndiceITBGL.aspx",
                           "https://web.bevsa.com.uy/CurvasVectorPrecios/CurvasIndices/Indices/IndiceINDUI.aspx",
                           "https://web.bevsa.com.uy/CurvasVectorPrecios/CurvasIndices/Indices/IndiceITLUP.aspx"],
                       "provider": ["BEVSA"]}},
        "deposits":
        {"dl": {
            "main": "https://www.bcu.gub.uy/Servicios-Financieros-SSF/Series%20IF/Depositos.xlsx"},
            "source": {"direct": ["https://www.bcu.gub.uy/Servicios-Financieros-SSF/Series%20IF/Depositos.xls"],
                       "indirect": [
                           "https://www.bcu.gub.uy/Servicios-Financieros-SSF/Paginas/Series-estadisticas-Depositos.aspx"],
                       "provider": ["BCU"]}},
        "credit":
        {"dl": {
            "main": "https://www.bcu.gub.uy/Servicios-Financieros-SSF/Series%20IF/Creditos.xlsx "},
            "source": {"direct": ["https://www.bcu.gub.uy/Servicios-Financieros-SSF/Series%20IF/Creditos.xls"],
                       "indirect": [
                           "https://www.bcu.gub.uy/Servicios-Financieros-SSF/Paginas/Series-Estadisticas-Creditos.aspx"],
                       "provider": ["BCU"]}},
        "interest_rates":
        {"dl": {
            "main": "https://www.bcu.gub.uy/Servicios-Financieros-SSF/Series%20IF/tasas.xls"},
            "source": {"direct": ["https://www.bcu.gub.uy/Servicios-Financieros-SSF/Series%20IF/tasas.xls"],
                       "indirect": [
                           "https://www.bcu.gub.uy/Servicios-Financieros-SSF/Paginas/Series-Estadisticas-Tasas.aspx"],
                       "provider": ["BCU"]}},
        "cpi":
        {"dl": {
            "main": "http://ine.gub.uy/c/document_library/get_file?uuid=2e92084a-94ec-4fec-b5ca-42b40d5d2826&groupId=10181"},
            "source": {"direct": [
                "http://ine.gub.uy/c/document_library/get_file?uuid=2e92084a-94ec-4fec-b5ca-42b40d5d2826&groupId=10181"],
                "indirect": [
                    "http://www.ine.gub.uy/web/guest/ipc-indice-de-precios-del-consumo"],
                "provider": ["INE"]}},
        "nxr_monthly":
        {"dl": {
            "main": "http://ine.gub.uy/c/document_library/get_file?uuid=3fbf4ffd-a829-420c-aca9-9f01ecd7919a&groupId=10181"},
            "source": {"direct": [
                "http://ine.gub.uy/c/document_library/get_file?uuid=3fbf4ffd-a829-420c-aca9-9f01ecd7919a&groupId=10181"],
                "indirect": [
                    "http://www.ine.gub.uy/web/guest/cotizacion-de-monedas2"],
                "provider": ["INE"]}},
        "nxr_daily":
        {"dl": {
            "main": "https://www.bcu.gub.uy/_layouts/15/BCU.Cotizaciones/handler/FileHandler.ashx?op=downloadcotizacionesexcel&KeyValuePairs={%22KeyValuePairs%22:{%22Monedas%22:[{%22Val%22:%222224%22,%22Text%22:%22DLS.%20USA%20CABLE%22}],"},
            "source": {"direct": [],
                       "indirect": [
                           "https://www.bcu.gub.uy/Estadisticas-e-Indicadores/Paginas/Cotizaciones.aspx"],
                       "provider": ["BCU"]}},
        "commodity_prices":
        {"dl": {
            "beef": "https://www.inac.uy/innovaportal/v/9799/10/innova.front/serie-semanal-ingreso-medio-de-exportacion---bovino-ovino-y-otros-productos",
            "pulp": f"https://www.insee.fr/en/statistiques/serie/telecharger/csv/010600339?ordre="
                    f"antechronologique&transposition=donneescolonne&periodeDebut=1&anneeDebut=1990&periodeFin=4&anneeFin="
                    f"{dt.datetime.now().year}",
            "soybean": "https://www.quandl.com/api/v3/datasets/CHRIS/CME_S1.csv?api_key=3TPxACcrxy9WsE871Lqe",
            "wheat": "https://www.quandl.com/api/v3/datasets/CHRIS/CME_W1.csv?api_key=3TPxACcrxy9WsE871Lqe",
            "milk1": "https://www.inale.org/estadisticas/",
            "milk2": "https://ec.europa.eu/info/sites/info/files/food-farming-fisheries/farming/documents/eu-milk-historical-price-series_en.xls",
            "imf": "https://www.imf.org/en/Research/commodity-prices"},
            "source": {"direct": [
                "https://www.inac.uy/innovaportal/v/9799/10/innova.front/serie-semanal-ingreso-medio-de-exportacion---bovino-ovino-y-otros-productos",
                f"https://www.insee.fr/en/statistiques/serie/telecharger/csv/010600339?ordre=antechronologique&"
                f"transposition=donneescolonne&periodeDebut=1&anneeDebut=1990&periodeFin=3&anneeFin="
                f"{dt.datetime.now().year}",
                "https://www.quandl.com/api/v3/datasets/CHRIS/CME_S1.csv?api_key=3TPxACcrxy9WsE871Lqe",
                "https://www.quandl.com/api/v3/datasets/CHRIS/CME_W1.csv?api_key=3TPxACcrxy9WsE871Lqe",
                "https://ec.europa.eu/info/sites/info/files/food-farming-fisheries/farming/documents/eu-milk-historical-price-series_en.xls"],
                "indirect": [
                    "https://www.inac.uy/innovaportal/v/5541/10/innova.front/precios",
                    "https://www.insee.fr/fr/statistiques/serie/010600339",
                    "https://www.quandl.com/data/CHRIS/CME_S1-Soybean-Futures-Continuous-Contract-1-S1-Front-Month",
                    "https://www.quandl.com/data/CHRIS/CME_W1-Wheat-Futures-Continuous-Contract-1-W1-Front-Month",
                    "https://www.inale.org/estadisticas/",
                    "https://ec.europa.eu/info/food-farming-fisheries/farming/facts-and-figures/markets/overviews/market-observatories/milk",
                    "https://www.imf.org/en/Research/commodity-prices"],
                "provider": [
                    "econuy en base a INAC, INSEE, Quandl, INALE, Comisión Europea y FMI"]}},
        "commodity_index":
        {"dl": {},
            "source": {"direct": [
                "https://www.inac.uy/innovaportal/v/9799/10/innova.front/serie-semanal-ingreso-medio-de-exportacion---bovino-ovino-y-otros-productos",
                f"https://www.insee.fr/en/statistiques/serie/telecharger/010600339?ordre=antechronologique&"
                f"transposition=donneescolonne&periodeDebut=1&anneeDebut=1990&periodeFin=11&anneeFin="
                f"{dt.datetime.now().year}",
                "https://www.quandl.com/api/v3/datasets/CHRIS/CME_S1.csv?api_key=3TPxACcrxy9WsE871Lqe",
                "https://www.quandl.com/api/v3/datasets/CHRIS/CME_W1.csv?api_key=3TPxACcrxy9WsE871Lqe",
                "https://ec.europa.eu/info/sites/info/files/food-farming-fisheries/farming/documents/eu-milk-historical-price-series_en.xls"],
                "indirect": [
                    "https://www.inac.uy/innovaportal/v/5541/10/innova.front/precios",
                    "https://www.insee.fr/fr/statistiques/serie/010600339",
                    "https://www.quandl.com/data/CHRIS/CME_S1-Soybean-Futures-Continuous-Contract-1-S1-Front-Month",
                    "https://www.quandl.com/data/CHRIS/CME_W1-Wheat-Futures-Continuous-Contract-1-W1-Front-Month",
                    "https://www.inale.org/estadisticas/",
                    "https://ec.europa.eu/info/food-farming-fisheries/farming/facts-and-figures/markets/overviews/market-observatories/milk",
                    "https://www.imf.org/en/Research/commodity-prices",
                    "https://comtrade.un.org/"],
                "provider": [
                    "econuy en base a INAC, INSEE, Quandl, INALE, Comisión Europea, FMI y Naciones Unidas"]}},
        "balance_gps": {"dl": {
            "main": "https://www.gub.uy/ministerio-economia-finanzas/datos-y-estadisticas/datos/informacion-resultados-del-sector-publico"},
            "source": {"direct": [],
                       "indirect": [
                       "https://www.gub.uy/ministerio-economia-finanzas/datos-y-estadisticas/datos/informacion-resultados-del-sector-publico"],
                       "provider": ["MEF"]}},
        "balance_nfps": {"dl": {
            "main": "https://www.gub.uy/ministerio-economia-finanzas/datos-y-estadisticas/datos/informacion-resultados-del-sector-publico"},
            "source": {"direct": [],
                       "indirect": [
                       "https://www.gub.uy/ministerio-economia-finanzas/datos-y-estadisticas/datos/informacion-resultados-del-sector-publico"],
                       "provider": ["MEF"]}},
        "balance_cg-bps": {"dl": {
            "main": "https://www.gub.uy/ministerio-economia-finanzas/datos-y-estadisticas/datos/informacion-resultados-del-sector-publico"},
            "source": {"direct": [],
                       "indirect": [
                       "https://www.gub.uy/ministerio-economia-finanzas/datos-y-estadisticas/datos/informacion-resultados-del-sector-publico"],
                       "provider": ["MEF"]}},
        "balance_pe": {"dl": {
            "main": "https://www.gub.uy/ministerio-economia-finanzas/datos-y-estadisticas/datos/informacion-resultados-del-sector-publico"},
            "source": {"direct": [],
                       "indirect": [
                       "https://www.gub.uy/ministerio-economia-finanzas/datos-y-estadisticas/datos/informacion-resultados-del-sector-publico"],
                       "provider": ["MEF"]}},
        "balance_ancap": {"dl": {
            "main": "https://www.gub.uy/ministerio-economia-finanzas/datos-y-estadisticas/datos/informacion-resultados-del-sector-publico"},
            "source": {"direct": [],
                       "indirect": [
                       "https://www.gub.uy/ministerio-economia-finanzas/datos-y-estadisticas/datos/informacion-resultados-del-sector-publico"],
                       "provider": ["MEF"]}},
        "balance_ute": {"dl": {
            "main": "https://www.gub.uy/ministerio-economia-finanzas/datos-y-estadisticas/datos/informacion-resultados-del-sector-publico"},
            "source": {"direct": [],
                       "indirect": [
                       "https://www.gub.uy/ministerio-economia-finanzas/datos-y-estadisticas/datos/informacion-resultados-del-sector-publico"],
                       "provider": ["MEF"]}},
        "balance_antel": {"dl": {
            "main": "https://www.gub.uy/ministerio-economia-finanzas/datos-y-estadisticas/datos/informacion-resultados-del-sector-publico"},
            "source": {"direct": [],
                       "indirect": [
                       "https://www.gub.uy/ministerio-economia-finanzas/datos-y-estadisticas/datos/informacion-resultados-del-sector-publico"],
                       "provider": ["MEF"]}},
        "balance_ose": {"dl": {
            "main": "https://www.gub.uy/ministerio-economia-finanzas/datos-y-estadisticas/datos/informacion-resultados-del-sector-publico"},
            "source": {"direct": [],
                       "indirect": [
                       "https://www.gub.uy/ministerio-economia-finanzas/datos-y-estadisticas/datos/informacion-resultados-del-sector-publico"],
                       "provider": ["MEF"]}},
        "balance_summary": {"dl": {
            "main": "https://www.gub.uy/ministerio-economia-finanzas/datos-y-estadisticas/datos/informacion-resultados-del-sector-publico"},
            "source": {"direct": [],
                       "indirect": [
                       "https://www.gub.uy/ministerio-economia-finanzas/datos-y-estadisticas/datos/informacion-resultados-del-sector-publico"],
                       "provider": ["MEF"]}},
        "tax_revenue": {"dl": {
            "main": "https://www.dgi.gub.uy/wdgi/afiledownload?2,4,865,O,S,0,19353%3BS%3B100%3B108,",
            "pdfs": "https://raw.githubusercontent.com/rxavier/econuy-extras/main/econuy_extras/retrieval/taxes_pdf.csv"},
            "source": {"direct": ["https://www.dgi.gub.uy/wdgi/afiledownload?2,4,865,O,S,0,19353%3BS%3B100%3B108,"],
                       "indirect": [
                       "https://www.dgi.gub.uy/wdgi/page?2,principal,dgi--datos-y-series-estadisticas--serie-de-datos--recaudacion-anual-y-mensual-por-impuesto,O,es,0,",
                       f"https://www.dgi.gub.uy/wdgi/page?2,principal,dgi--datos-y-series-estadisticas--informes-mensuales-de-la-recaudacion-{dt.datetime.now().year},O,es,0,"],
                       "provider": ["DGI"]}},
        "public_debt_gps": {"dl": {
            "main": "https://www.bcu.gub.uy/Estadisticas-e-Indicadores/Endeudamiento%20Pblico/resdspg.xls"},
            "source": {"direct": [
                "https://www.bcu.gub.uy/Estadisticas-e-Indicadores/Endeudamiento%20Pblico/resdspg.xls"],
            "indirect": [
                "https://www.bcu.gub.uy/Estadisticas-e-Indicadores/Paginas/Default.aspx"],
            "provider": ["BCU"]}},
        "public_debt_nfps": {"dl": {
            "main": "https://www.bcu.gub.uy/Estadisticas-e-Indicadores/Endeudamiento%20Pblico/resdspg.xls"},
            "source": {"direct": [
                "https://www.bcu.gub.uy/Estadisticas-e-Indicadores/Endeudamiento%20Pblico/resdspg.xls"],
            "indirect": [
                "https://www.bcu.gub.uy/Estadisticas-e-Indicadores/Paginas/Default.aspx"],
            "provider": ["BCU"]}},
        "public_debt_cb": {"dl": {
            "main": "https://www.bcu.gub.uy/Estadisticas-e-Indicadores/Endeudamiento%20Pblico/resdspg.xls"},
            "source": {"direct": [
                "https://www.bcu.gub.uy/Estadisticas-e-Indicadores/Endeudamiento%20Pblico/resdspg.xls"],
            "indirect": [
                "https://www.bcu.gub.uy/Estadisticas-e-Indicadores/Paginas/Default.aspx"],
            "provider": ["BCU"]}},
        "net_public_debt": {"dl": {},
                            "source": {"direct": [
                                "https://www.bcu.gub.uy/Estadisticas-e-Indicadores/Endeudamiento%20Pblico/resdspg.xls",
                                "https://www.bcu.gub.uy/Estadisticas-e-Indicadores/MonedayCredito/Activos-de-Reserva/reservas.xls"],
            "indirect": [
                                "https://www.bcu.gub.uy/Estadisticas-e-Indicadores/Paginas/Default.aspx"],
                            "provider": [
                                "econuy en base a BCU"]}},
        "diesel": {"dl": {
            "main": "https://www.gub.uy/ministerio-industria-energia-mineria/datos-y-estadisticas/datos/series-estadisticas-petroleo-derivados"},
            "source": {"direct": [],
                       "indirect": [
                       "https://www.gub.uy/ministerio-industria-energia-mineria/datos-y-estadisticas/datos/series-estadisticas-petroleo-derivados"],
                       "provider": ["MIEM"]}},
        "gasoline": {"dl": {
            "main": "https://www.gub.uy/ministerio-industria-energia-mineria/datos-y-estadisticas/datos/series-estadisticas-petroleo-derivados"},
            "source": {"direct": [],
                       "indirect": [
                       "https://www.gub.uy/ministerio-industria-energia-mineria/datos-y-estadisticas/datos/series-estadisticas-petroleo-derivados"],
                       "provider": ["MIEM"]}},
        "electricity": {"dl": {
            "main": "https://www.gub.uy/ministerio-industria-energia-mineria/datos-y-estadisticas/datos/series-estadisticas-energia-electrica"},
            "source": {"direct": [],
                       "indirect": [
                       "https://www.gub.uy/ministerio-industria-energia-mineria/datos-y-estadisticas/datos/series-estadisticas-energia-electrica"],
                       "provider": ["MIEM"]}},
        "labor_rates": {"dl": {
            "main": "http://ine.gub.uy/c/document_library/get_file?uuid=50ae926c-1ddc-4409-afc6-1fecf641e3d0&groupId=10181",
            "missing": "https://docs.google.com/spreadsheets/d/1amqU3fUSok0kDB_LYvCZlBtVpkZeTK5vW5O8bShlOtw/export?format=xlsx&authuser=0"},
            "source": {"direct": [
                "http://ine.gub.uy/c/document_library/get_file?uuid=50ae926c-1ddc-4409-afc6-1fecf641e3d0&groupId=10181"],
            "indirect": [
                "http://www.ine.gub.uy/web/guest/actividad-empleo-y-desempleo"],
            "provider": ["INE"]}},
        "nominal_wages": {"dl": {
            "historical": "http://www.ine.gub.uy/c/document_library/get_file?uuid=a76433b7-5fba-40fc-9958-dd913338e989&groupId=10181",
            "current": "http://www.ine.gub.uy/c/document_library/get_file?uuid=97f07fd8-9410-476e-bf81-e6b1c11467ef&groupId=10181"},
            "source": {"direct": [
                "http://www.ine.gub.uy/c/document_library/get_file?uuid=a76433b7-5fba-40fc-9958-dd913338e989&groupId=10181",
                "http://www.ine.gub.uy/c/document_library/get_file?uuid=97f07fd8-9410-476e-bf81-e6b1c11467ef&groupId=10181"],
            "indirect": [
                "http://www.ine.gub.uy/web/guest/ims-indice-medio-de-salarios"],
            "provider": ["INE"]}},
        "hours_worked": {"dl": {
            "main": "http://www.ine.gub.uy/c/document_library/get_file?uuid=167e0db0-95ca-45d2-8e81-3b8c5bb8f9ee&groupId=10181",
            "historical": "http://www.ine.gub.uy/c/document_library/get_file?uuid=73ac6ede-8452-48b8-ad32-993d3b047091&groupId=10181",
            "missing": "https://docs.google.com/spreadsheets/d/1amqU3fUSok0kDB_LYvCZlBtVpkZeTK5vW5O8bShlOtw/export?format=xlsx&authuser=0"},
            "source": {"direct": [
                "http://www.ine.gub.uy/c/document_library/get_file?uuid=73ac6ede-8452-48b8-ad32-993d3b047091&groupId=10181",
                "http://www.ine.gub.uy/c/document_library/get_file?uuid=167e0db0-95ca-45d2-8e81-3b8c5bb8f9ee&groupId=10181"],
            "indirect": [
                "http://www.ine.gub.uy/web/guest/actividad-empleo-y-desempleo"],
            "provider": ["INE"]}},
        "income_household": {"dl": {
            "main": "http://www.ine.gub.uy/c/document_library/get_file?uuid=40bd0267-3922-478d-8bc0-252f508a72fe&groupId=10181",
            "missing": "https://docs.google.com/spreadsheets/d/1amqU3fUSok0kDB_LYvCZlBtVpkZeTK5vW5O8bShlOtw/export?format=xlsx&authuser=0"},
            "source": {"direct": [
                "http://www.ine.gub.uy/c/document_library/get_file?uuid=40bd0267-3922-478d-8bc0-252f508a72fe&groupId=10181"],
            "indirect": [
                "http://www.ine.gub.uy/web/guest/gastos-e-ingresos-de-las-personas-y-los-hogares"],
            "provider": ["INE"]}},
        "income_capita": {"dl": {
            "main": "http://www.ine.gub.uy/c/document_library/get_file?uuid=ca57dafa-8091-4c2f-8df8-7b8445859b93&groupId=10181",
            "missing": "https://docs.google.com/spreadsheets/d/1amqU3fUSok0kDB_LYvCZlBtVpkZeTK5vW5O8bShlOtw/export?format=xlsx&authuser=0"},
            "source": {"direct": [
                "http://www.ine.gub.uy/c/document_library/get_file?uuid=ca57dafa-8091-4c2f-8df8-7b8445859b93&groupId=10181"],
            "indirect": [
                "http://www.ine.gub.uy/web/guest/gastos-e-ingresos-de-las-personas-y-los-hogares"],
            "provider": ["INE"]}},
        "natacc_ind_con_nsa": {
            "dl": {"main": "https://www.bcu.gub.uy/Estadisticas-e-Indicadores/Cuentas%20Nacionales/cuadro_101t.xls"},
            "source": {"direct": ["https://www.bcu.gub.uy/Estadisticas-e-Indicadores/Cuentas%20Nacionales/cuadro_101t.xls"],
                       "indirect": [
                       "https://www.bcu.gub.uy/Estadisticas-e-Indicadores/Cuentas%20Nacionales/presentacion05t.htm"],
                       "provider": ["BCU"]}},
        "natacc_ind_cur_nsa": {
            "dl": {"main": "https://www.bcu.gub.uy/Estadisticas-e-Indicadores/Cuentas%20Nacionales/cuadro_100t.xls"},
            "source": {"direct": ["https://www.bcu.gub.uy/Estadisticas-e-Indicadores/Cuentas%20Nacionales/cuadro_100t.xls"],
                       "indirect": [
                       "https://www.bcu.gub.uy/Estadisticas-e-Indicadores/Cuentas%20Nacionales/presentacion05t.htm"],
                       "provider": ["BCU"]}},
        "natacc_gas_con_nsa": {
            "dl": {"main": "https://www.bcu.gub.uy/Estadisticas-e-Indicadores/Cuentas%20Nacionales/cuadro_104t.xls"},
            "source": {"direct": ["https://www.bcu.gub.uy/Estadisticas-e-Indicadores/Cuentas%20Nacionales/cuadro_104t.xls"],
                       "indirect": [
                       "https://www.bcu.gub.uy/Estadisticas-e-Indicadores/Cuentas%20Nacionales/presentacion05t.htm"],
                       "provider": ["BCU"]}},
        "natacc_ind_con_idx_nsa": {
            "dl": {"main": "https://www.bcu.gub.uy/Estadisticas-e-Indicadores/Cuentas%20Nacionales/cuadro_132t.xls"},
            "source": {"direct": ["https://www.bcu.gub.uy/Estadisticas-e-Indicadores/Cuentas%20Nacionales/cuadro_132t.xls"],
                       "indirect": [
                       "https://www.bcu.gub.uy/Estadisticas-e-Indicadores/Cuentas%20Nacionales/presentacion05t.htm"],
                       "provider": ["BCU"]}},
        "natacc_ind_con_idx_sa": {
            "dl": {"main": "https://www.bcu.gub.uy/Estadisticas-e-Indicadores/Cuentas%20Nacionales/cuadro_133t.xls"},
            "source": {"direct": ["https://www.bcu.gub.uy/Estadisticas-e-Indicadores/Cuentas%20Nacionales/cuadro_133t.xls"],
                       "indirect": [
                       "https://www.bcu.gub.uy/Estadisticas-e-Indicadores/Cuentas%20Nacionales/presentacion05t.htm"],
                       "provider": ["BCU"]}},
        "natacc_gdp_cur_nsa": {
            "dl": {"main": "https://www.bcu.gub.uy/Estadisticas-e-Indicadores/Cuentas%20Nacionales/cuadro_130t.xls"},
            "source": {"direct": ["https://www.bcu.gub.uy/Estadisticas-e-Indicadores/Cuentas%20Nacionales/cuadro_130t.xls"],
                       "indirect": [
                       "https://www.bcu.gub.uy/Estadisticas-e-Indicadores/Cuentas%20Nacionales/presentacion05t.htm"],
                       "provider": ["BCU"]}},
        "reserves": {"dl": {
            "main": "https://www.bcu.gub.uy/Estadisticas-e-Indicadores/MonedayCredito/Activos-de-Reserva/reservas.xls"},
            "source": {"direct": [
                "https://www.bcu.gub.uy/Estadisticas-e-Indicadores/MonedayCredito/Activos-de-Reserva/reservas.xls"],
            "indirect": [
                "https://www.bcu.gub.uy/Estadisticas-e-Indicadores/Paginas/Default.aspx"],
            "provider": ["BCU"]}},
        "industrial_production": {"dl": {
            "main": "http://www.ine.gub.uy/c/document_library/get_file?uuid=8e08c0dc-acc2-44f7-b302-daa32e0b978b&groupId=10181",
            "weights": "http://www.ine.gub.uy/c/document_library/get_file?uuid=864b4bc2-626d-49ab-82ef-1bcf08360da1&groupId=10181"},
            "source": {"direct": [
                "http://www.ine.gub.uy/c/document_library/get_file?uuid=8e08c0dc-acc2-44f7-b302-daa32e0b978b&groupId=10181"],
            "indirect": [
                "http://www.ine.gub.uy/web/guest/industria-manufacturera"],
            "provider": ["INE"]}},
        "cattle": {"dl": {
            "main": "https://www.inac.uy/innovaportal/v/11998/10/innova.front/serie-semanal-faena---bovinos-y-ovinos"},
            "source": {"direct": [
                "https://www.inac.uy/innovaportal/v/11998/10/innova.front/serie-semanal-faena---bovinos-y-ovinos"],
            "indirect": [
                "https://www.inac.uy/innovaportal/v/5539/10/innova.front/faena"],
            "provider": ["INAC"]}},
        "milk": {"dl": {
            "main": "https://www.inale.org/estadisticas/remision-a-planta/"},
            "source": {"direct": [],
                       "indirect": [
                       "https://www.inale.org/estadisticas/remision-a-planta/"],
                       "provider": ["INALE"]}},
        "cement": {"dl": {
            "main": "http://www.ciu.com.uy/innovaportal/file/83062/1/cemento-web.xlsx"},
            "source": {"direct": ["http://www.ciu.com.uy/innovaportal/file/83062/1/cemento-web.xlsx"],
                       "indirect": [
                       "http://www.ciu.com.uy/innovaportal/v/83062/9/innova.front/series-de-ventas-mensuales-por-destino.html"],
                       "provider": ["AFCPU y CIU"]}},
        "consumer_confidence": {"dl": {
            "main": "https://ucu.edu.uy/sites/default/files/facultad/fce/i_competitividad/serie_icc_-.xlsx"},
            "source": {"direct": ["https://ucu.edu.uy/sites/default/files/facultad/fce/i_competitividad/serie_icc_-.xlsx"],
                       "indirect": [
                "https://ucu.edu.uy/es/icc"],
            "provider": ["UCU"]}},
        "sovereign_risk": {"dl": {
            "historical": "https://www4.rafap.com.uy/internet/images/indicadores/UBI_Historico.xls",
            "current": "https://www4.rafap.com.uy/internet/servlet/hextubicd"},
            "source": {"direct": [],
                       "indirect": [
                       "https://www.rafap.com.uy/mvdcms/Institucional/UBI-uc89"],
                       "provider": ["República AFAP"]}},
        "reserves_changes": {"dl": {
            "main": "https://www.bcu.gub.uy/Estadisticas-e-Indicadores/Informe%20Diario%20Pasivos%20Monetarios/infd_",
            "missing": "https://docs.google.com/spreadsheets/d/1tXwv8SaigbBrfBSSCVGBjSs88f3dgTq4nIANPn7vjYI/export?format=xlsx&authuser=0"},
            "source": {"direct": [],
                       "indirect": [
                       "https://www.bcu.gub.uy/Estadisticas-e-Indicadores/Paginas/Informe-Diario-Pasivos-Monetarios.aspx"],
                       "provider": ["BCU"]}},
        "rxr_official": {"dl": {
            "main": "https://www.bcu.gub.uy/Estadisticas-e-Indicadores/Indice_Cambio_Real/TCRE.xls"},
            "source": {"direct": ["https://www.bcu.gub.uy/Estadisticas-e-Indicadores/Indice_Cambio_Real/TCRE.xls"],
                       "indirect": [
                       "https://www.bcu.gub.uy/Estadisticas-e-Indicadores/Paginas/Cambio-Real-Efectivo.aspx"],
                       "provider": ["BCU"]}},
        "rxr_custom": {"dl": {},
                       "source": {"direct": [],
                                  "indirect": [
                                  "https://data.imf.org/?sk=4c514d48-b6ba-49ed-8ab9-52b0c1a0179b",
                                  "http://www.ine.gub.uy/web/guest/ipc-indice-de-precios-del-consumo",
                                  "http://www.ine.gub.uy/web/guest/cotizacion-de-monedas2",
                                  "https://www.ambito.com/contenidos/dolar-informal.html",
                                  "http://www.bcra.gov.ar/PublicacionesEstadisticas/Principales_variables_datos.asp?serie=7931&detalle=Inflaci%F3n%20mensual%A0(variaci%F3n%20en%20%)",
                                  "http://www.inflacionverdadera.com/argentina/",
                                  "http://www.ipeadata.gov.br/Default.aspx",
                                  "https://www.ibge.gov.br/estatisticas/economicas/precos-e-custos/9256-indice-nacional-de-precos-ao-consumidor-amplo.html?=&t=o-que-e",
                       ],
            "provider": [
                           "econuy en base a INE, BCRA, Ipea, IBGE, FMI, Ámbito e Inflación Verdadera"]}},
        "trade_x_prod_val": {
            "dl": {"main": "https://www.bcu.gub.uy/Estadisticas-e-Indicadores/ComercioExterior_ICB/exp_ciiu_val.xls"},
            "source": {
                "direct": ["https://www.bcu.gub.uy/Estadisticas-e-Indicadores/ComercioExterior_ICB/exp_ciiu_val.xls"],
                "indirect": [
                    "https://www.bcu.gub.uy/Estadisticas-e-Indicadores/Paginas/Intercambio-Comercial-.aspx"],
                "provider": ["BCU"]}},
        "trade_x_prod_vol": {
            "dl": {"main": "https://www.bcu.gub.uy/Estadisticas-e-Indicadores/ComercioExterior_ICB/web_exp_ciiu_ivf.xls"},
            "source": {
                "direct": ["https://www.bcu.gub.uy/Estadisticas-e-Indicadores/ComercioExterior_ICB/web_exp_ciiu_ivf.xls"],
                "indirect": [
                    "https://www.bcu.gub.uy/Estadisticas-e-Indicadores/Paginas/Intercambio-Comercial-.aspx"],
                "provider": ["BCU"]}},
        "trade_x_prod_pri": {
            "dl": {"main": "https://www.bcu.gub.uy/Estadisticas-e-Indicadores/ComercioExterior_ICB/web_exp_ciiu_ip.xls"},
            "source": {
                "direct": ["https://www.bcu.gub.uy/Estadisticas-e-Indicadores/ComercioExterior_ICB/web_exp_ciiu_ip.xls"],
                "indirect": [
                    "https://www.bcu.gub.uy/Estadisticas-e-Indicadores/Paginas/Intercambio-Comercial-.aspx"],
                "provider": ["BCU"]}},
        "trade_x_dest_val": {
            "dl": {"main": "https://www.bcu.gub.uy/Estadisticas-e-Indicadores/ComercioExterior_ICB/exp_pais_val.xls"},
            "source": {
                "direct": ["https://www.bcu.gub.uy/Estadisticas-e-Indicadores/ComercioExterior_ICB/exp_pais_val.xls"],
                "indirect": [
                    "https://www.bcu.gub.uy/Estadisticas-e-Indicadores/Paginas/Intercambio-Comercial-.aspx"],
                "provider": ["BCU"]}},
        "trade_x_dest_vol": {
            "dl": {"main": "https://www.bcu.gub.uy/Estadisticas-e-Indicadores/ComercioExterior_ICB/web_exp_pais_ivf.xls"},
            "source": {
                "direct": ["https://www.bcu.gub.uy/Estadisticas-e-Indicadores/ComercioExterior_ICB/web_exp_pais_ivf.xls"],
                "indirect": [
                    "https://www.bcu.gub.uy/Estadisticas-e-Indicadores/Paginas/Intercambio-Comercial-.aspx"],
                "provider": ["BCU"]}},
        "trade_x_dest_pri": {
            "dl": {"main": "https://www.bcu.gub.uy/Estadisticas-e-Indicadores/ComercioExterior_ICB/web_exp_pais_ip.xls"},
            "source": {
                "direct": ["https://www.bcu.gub.uy/Estadisticas-e-Indicadores/ComercioExterior_ICB/web_exp_pais_ip.xls"],
                "indirect": [
                    "https://www.bcu.gub.uy/Estadisticas-e-Indicadores/Paginas/Intercambio-Comercial-.aspx"],
                "provider": ["BCU"]}},
        "trade_m_sect_val": {
            "dl": {"main": "https://www.bcu.gub.uy/Estadisticas-e-Indicadores/ComercioExterior_ICB/imp_gce_val.xls"},
            "source": {
                "direct": ["https://www.bcu.gub.uy/Estadisticas-e-Indicadores/ComercioExterior_ICB/imp_gce_val.xls"],
                "indirect": [
                    "https://www.bcu.gub.uy/Estadisticas-e-Indicadores/Paginas/Intercambio-Comercial-.aspx"],
                "provider": ["BCU"]}},
        "trade_m_sect_vol": {
            "dl": {"main": "https://www.bcu.gub.uy/Estadisticas-e-Indicadores/ComercioExterior_ICB/web_imp_gce_ivf.xls"},
            "source": {
                "direct": ["https://www.bcu.gub.uy/Estadisticas-e-Indicadores/ComercioExterior_ICB/web_imp_gce_ivf.xls"],
                "indirect": [
                    "https://www.bcu.gub.uy/Estadisticas-e-Indicadores/Paginas/Intercambio-Comercial-.aspx"],
                "provider": ["BCU"]}},
        "trade_m_sect_pri": {
            "dl": {"main": "https://www.bcu.gub.uy/Estadisticas-e-Indicadores/ComercioExterior_ICB/web_imp_gce_ip.xls"},
            "source": {
                "direct": ["https://www.bcu.gub.uy/Estadisticas-e-Indicadores/ComercioExterior_ICB/web_imp_gce_ip.xls"],
                "indirect": [
                    "https://www.bcu.gub.uy/Estadisticas-e-Indicadores/Paginas/Intercambio-Comercial-.aspx"],
                "provider": ["BCU"]}},
        "trade_m_orig_val": {
            "dl": {"main": "https://www.bcu.gub.uy/Estadisticas-e-Indicadores/ComercioExterior_ICB/imp_pais_val.xls"},
            "source": {
                "direct": ["https://www.bcu.gub.uy/Estadisticas-e-Indicadores/ComercioExterior_ICB/imp_pais_val.xls"],
                "indirect": [
                    "https://www.bcu.gub.uy/Estadisticas-e-Indicadores/Paginas/Intercambio-Comercial-.aspx"],
                "provider": ["BCU"]}},
        "trade_m_orig_vol": {
            "dl": {"main": "https://www.bcu.gub.uy/Estadisticas-e-Indicadores/ComercioExterior_ICB/web_imp_pais_ivf.xls"},
            "source": {
                "direct": ["https://www.bcu.gub.uy/Estadisticas-e-Indicadores/ComercioExterior_ICB/web_imp_pais_ivf.xls"],
                "indirect": [
                    "https://www.bcu.gub.uy/Estadisticas-e-Indicadores/Paginas/Intercambio-Comercial-.aspx"],
                "provider": ["BCU"]}},
        "trade_m_orig_pri": {
            "dl": {"main": "https://www.bcu.gub.uy/Estadisticas-e-Indicadores/ComercioExterior_ICB/web_imp_pais_ip.xls"},
            "source": {
                "direct": ["https://www.bcu.gub.uy/Estadisticas-e-Indicadores/ComercioExterior_ICB/web_imp_pais_ip.xls"],
                "indirect": [
                    "https://www.bcu.gub.uy/Estadisticas-e-Indicadores/Paginas/Intercambio-Comercial-.aspx"],
                "provider": ["BCU"]}},
        "cpi_measures":
        {"dl": {
            "2010-14": "http://www.ine.gub.uy/c/document_library/get_file?uuid=668d4f77-74d8-46ba-8360-77bd867996df&groupId=10181",
            "2015-": "https://www.ine.gub.uy/c/document_library/get_file?uuid=ad969d52-cebc-4b40-9a1f-34ce277e463e&groupId=10181",
            "1997": "http://www.ine.gub.uy/c/document_library/get_file?uuid=1cd81500-420a-44d0-ae4d-0add9d913107&groupId=10181",
            "1997_weights": "https://docs.google.com/spreadsheets/d/1gSQdp6b97udmki0DZBndhLajLv6uGDX7kYb66BUswj8/export#gid=0"},
            "source": {"direct": [
                "http://www.ine.gub.uy/c/document_library/get_file?uuid=668d4f77-74d8-46ba-8360-77bd867996df&groupId=10181",
                "https://www.ine.gub.uy/c/document_library/get_file?uuid=ad969d52-cebc-4b40-9a1f-34ce277e463e&groupId=10181",
                "http://www.ine.gub.uy/c/document_library/get_file?uuid=1cd81500-420a-44d0-ae4d-0add9d913107&groupId=10181",
                "http://www.ine.gub.uy/c/document_library/get_file?uuid=5f2e75d2-5df6-48da-978d-e7930d47c037&groupId=10181"],
                "indirect": [
                    "http://www.ine.gub.uy/web/guest/ipc-indice-de-precios-del-consumo"],
                "provider": ["econuy en base a INE"]}},
        "balance_fss": {"dl": {},
                        "source": {"direct": [],
                                   "indirect": [
                                   "https://www.gub.uy/ministerio-economia-finanzas/datos-y-estadisticas/datos"],
                                   "provider": [
                                   "econuy en base a MEF"]}},
        "labor_rates_people": {"dl": {
            "act_5000": "http://www.ine.gub.uy/c/document_library/get_file?uuid=b51d8104-d367-4d0f-828b-189eefc29de2&groupId=10181",
            "emp_5000": "http://www.ine.gub.uy/c/document_library/get_file?uuid=0902797e-e588-4da3-91cd-153c4d1d28a5&groupId=10181",
            "des_5000": "http://www.ine.gub.uy/c/document_library/get_file?uuid=d1434567-3da4-4321-9341-4fb3d8b6a09c&groupId=10181",
            "population": "http://www.ine.gub.uy/c/document_library/get_file?uuid=2a5c1e6e-b02f-4a63-963f-925edea7c17e&groupId=10181"},
            "source": {"direct": [
                "http://ine.gub.uy/c/document_library/get_file?uuid=50ae926c-1ddc-4409-afc6-1fecf641e3d0&groupId=10181",
                "http://www.ine.gub.uy/c/document_library/get_file?uuid=b51d8104-d367-4d0f-828b-189eefc29de2&groupId=10181",
                "http://www.ine.gub.uy/c/document_library/get_file?uuid=0902797e-e588-4da3-91cd-153c4d1d28a5&groupId=10181",
                "http://www.ine.gub.uy/c/document_library/get_file?uuid=d1434567-3da4-4321-9341-4fb3d8b6a09c&groupId=10181",
                "http://www.ine.gub.uy/c/document_library/get_file?uuid=2a5c1e6e-b02f-4a63-963f-925edea7c17e&groupId=10181"],
            "indirect": [
                "http://www.ine.gub.uy/web/guest/actividad-empleo-y-desempleo",
                "http://www.ine.gub.uy/estimaciones-y-proyecciones"],
            "provider": [
                "econuy en base a INE"]}},
        "real_wages": {"dl": {},
                       "source": {"direct": [
                           "http://www.ine.gub.uy/c/document_library/get_file?uuid=a76433b7-5fba-40fc-9958-dd913338e989&groupId=10181",
                           "http://www.ine.gub.uy/c/document_library/get_file?uuid=97f07fd8-9410-476e-bf81-e6b1c11467ef&groupId=10181",
                           "http://www.ine.gub.uy/c/document_library/get_file?uuid=2e92084a-94ec-4fec-b5ca-42b40d5d2826&groupId=10181"],
            "indirect": [
                           "http://www.ine.gub.uy/web/guest/ims-indice-medio-de-salarios",
                           "http://www.ine.gub.uy/web/guest/ipc-indice-de-precios-del-consumo"],
                       "provider": [
                           "econuy en base a INE"]}},
        "trade_balance": {"dl": {},
                          "source": {"direct": [
                              "https://www.bcu.gub.uy/Estadisticas-e-Indicadores/ComercioExterior_ICB/exp_pais_val.xls",
                              "https://www.bcu.gub.uy/Estadisticas-e-Indicadores/ComercioExterior_ICB/imp_pais_val.xls"],
            "indirect": [
                              "https://www.bcu.gub.uy/Estadisticas-e-Indicadores/Paginas/Intercambio-Comercial-.aspx"],
            "provider": [
                              "econuy en base a BCU"]}},
        "terms_of_trade": {"dl": {},
                           "source": {"direct": [
                               "https://www.bcu.gub.uy/Estadisticas-e-Indicadores/ComercioExterior_ICB/web_exp_pais_ip.xls",
                               "https://www.bcu.gub.uy/Estadisticas-e-Indicadores/ComercioExterior_ICB/web_imp_pais_ip.xls"],
            "indirect": [
                               "https://www.bcu.gub.uy/Estadisticas-e-Indicadores/Paginas/Intercambio-Comercial-.aspx"],
            "provider": [
                               "econuy en base a BCU"]}},
        "core_industrial": {"dl": {
            "main": "http://www.ine.gub.uy/c/document_library/get_file?uuid=8e08c0dc-acc2-44f7-b302-daa32e0b978b&groupId=10181",
            "weights": "http://www.ine.gub.uy/c/document_library/get_file?uuid=864b4bc2-626d-49ab-82ef-1bcf08360da1&groupId=10181"},
            "source": {"direct": [
                "http://www.ine.gub.uy/c/document_library/get_file?uuid=8e08c0dc-acc2-44f7-b302-daa32e0b978b&groupId=10181",
                "http://www.ine.gub.uy/c/document_library/get_file?uuid=1768051c-5b85-42ea-8340-b8b44ee2948c&groupId=10181"],
            "indirect": [
                "http://www.ine.gub.uy/web/guest/industria-manufacturera"],
            "provider": [
                "econuy en base a INE"]}},
        "global_gdp": {"dl": {
            "fred": "https://api.stlouisfed.org/fred/series/observations?series_id=",
            "chn_oecd": "https://stats.oecd.org/SDMX-JSON/data/QNA/CHN.B1_GE.GYSA+GPSA.Q/all?startTime=1960-Q1&endTime=",
            "chn_obs": "https://docs.google.com/spreadsheets/d/1JwHqYSyBCOj9E60X0JCnPIn4WEzfu6rcPN8Xg76AhyU/export?format=xlsx&authuser=0"},
            "source": {"direct": [],
                       "indirect": ["https://fred.stlouisfed.org/",
                                    "https://stats.oecd.org/Index.aspx",
                                    "https://data.stats.gov.cn/english/easyquery.htm?cn=B01"],
                       "provider": ["econuy en base a FRB St. Louis, OECD y NBS China"]}},
        "global_stocks": {"dl": {
            "spy": f"https://query1.finance.yahoo.com/v7/finance/download/%5EGSPC?period1=-1325635200&period2={dt.datetime.now().timestamp().__round__()}&interval=1d&events=history&includeAdjustedClose=true",
            "n100": f"https://query1.finance.yahoo.com/v7/finance/download/%5EN100?period1=946598400&period2={dt.datetime.now().timestamp().__round__()}&interval=1d&events=history&includeAdjustedClose=true",
            "sse": f"https://query1.finance.yahoo.com/v7/finance/download/000001.SS?period1=867715200&period2={dt.datetime.now().timestamp().__round__()}&interval=1d&events=history&includeAdjustedClose=true",
            "nikkei": f"https://query1.finance.yahoo.com/v7/finance/download/%5EN225?period1=-157507200&period2={dt.datetime.now().timestamp().__round__()}&interval=1d&events=history&includeAdjustedClose=true"},
            "source": {"direct": [],
                       "indirect": ["https://finance.yahoo.com/quote/%5EGSPC/history/",
                                    "https://finance.yahoo.com/quote/%5EN100/history/",
                                    "https://finance.yahoo.com/quote/000001.SS/history/",
                                    "https://finance.yahoo.com/quote/%5EN225/history/"],
                       "provider": ["Yahoo Finance"]}},
        "global_policy_rates": {"dl": {
            "main": "https://www.bis.org/statistics/full_webstats_cbpol_d_dataflow_csv_row.zip"},
            "source": {"direct": ["https://www.bis.org/statistics/full_webstats_cbpol_d_dataflow_csv_row.zip"],
                       "indirect": ["https://www.bis.org/statistics/cbpol.htm"],
                       "provider": ["BIS"]}},
        "global_long_rates": {"dl": {
            "main": "https://www.investing.com/instruments/HistoricalDataAjax",
            "fred": "https://api.stlouisfed.org/fred/series/observations?series_id="},
            "source": {"direct": [],
                       "indirect": ["https://fred.stlouisfed.org/series/DGS10",
                                    "https://www.investing.com/rates-bonds/government-bond-spreads"],
                       "provider": ["FRB St. Louis", "Investing.com"]}},
        "global_nxr": {"dl": {
            "dollar": f"https://query1.finance.yahoo.com/v7/finance/download/DX-Y.NYB?period1=31795200&period2={dt.datetime.now().timestamp().__round__()}&interval=1d&events=history&includeAdjustedClose=true",
            "eur": f"https://query1.finance.yahoo.com/v7/finance/download/USDEUR=X?period1=1070150400&period2={dt.datetime.now().timestamp().__round__()}&interval=1d&events=history&includeAdjustedClose=true",
            "jpy": f"https://query1.finance.yahoo.com/v7/finance/download/USDJPY=X?period1=846547200&period2={dt.datetime.now().timestamp().__round__()}&interval=1d&events=history&includeAdjustedClose=true",
            "cny": f"https://query1.finance.yahoo.com/v7/finance/download/USDCNY=X?period1=991180800&period2={dt.datetime.now().timestamp().__round__()}&interval=1d&events=history&includeAdjustedClose=true"},
            "source": {"direct": [],
                       "indirect": ["https://finance.yahoo.com/quote/DX-Y.NYB?/history/",
                                    "https://finance.yahoo.com/quote/USDEUR=X/history/",
                                    "https://finance.yahoo.com/quote/USDJPY=X/history/",
                                    "https://finance.yahoo.com/quote/USDCNY=X/history/"],
                       "provider": ["Yahoo Finance"]}},
        "regional_embi_spreads": {"dl": {
            "global": "https://cdn.bancentral.gov.do/documents/entorno-internacional/documents/Serie_Historica_Spread_del_EMBI.xlsx",
            "brasil": f"https://mercados.ambito.com//riesgopaisinternacional/brasil/historico-general/01-01-1990/{dt.datetime.now().strftime('%d-%m-%Y')}",
            "argentina": f"https://mercados.ambito.com//riesgopais/historico-general/01-01-1990/{dt.datetime.now().strftime('%d-%m-%Y')}"},
            "source": {"direct": [],
                       "indirect": ["https://www.ambito.com/contenidos/riesgo-pais.html",
                                    "https://www.bancentral.gov.do/a/d/2585-entorno-internacional"],
                       "provider": ["Ámbito, BCRD"]}},
        "regional_embi_yields": {"source": {"direct": [],
                                            "indirect": ["https://fred.stlouisfed.org/series/DGS10",
                                                         "https://www.ambito.com/contenidos/riesgo-pais.html",
                                                         "https://www.bancentral.gov.do/a/d/2585-entorno-internacional"],
                                            "provider": ["econuy en base a FRB St. Louis, Ámbito y BCRD"]}},
        "regional_gdp": {"dl": {"arg_new": "https://www.indec.gob.ar/indec/web/Nivel4-Tema-3-9-47",
                                "arg_old": "https://www.indec.gob.ar/ftp/nuevaweb/cuadros/17/cuadro12.xls",
                                "bra": "https://ftp.ibge.gov.br/Contas_Nacionais/Contas_Nacionais_Trimestrais/Tabelas_Completas/Tab_Compl_CNT.zip"},
                         "source": {"direct": [],
                                    "indirect": [
                                    "https://www.indec.gob.ar/indec/web/Nivel4-Tema-3-9-47",
                                    "https://www.indec.gob.ar/indec/web/Institucional-Indec-InformacionDeArchivo-5",
                                    "https://sidra.ibge.gov.br/tabela/1621#/n1/all/v/all/p/all/c11255/90687,90691,90696,90705,90707,93404,93405,93406,93407,93408/d/v584%201/l/v,c11255,t+p/resultado"],
                                    "provider": ["INDEC", "IBGE"]}},
        "regional_cpi": {"dl": {"ar": "http://www.bcra.gov.ar/PublicacionesEstadisticas/Principales_variables_datos.asp",
                                "ar_payload": f"fecha_desde=1970-01-01&fecha_hasta={dt.datetime.now().strftime('%Y-%m-%d')}&B1=Enviar&primeravez=1&fecha_desde=19600101&fecha_hasta={dt.datetime.now().strftime('%Y%m%d')}&serie=7931&serie1=0&serie2=0&serie3=0&serie4=0&detalle=Inflaci%F3n+mensual%A0%28variaci%F3n+en+%29",
                                "ar_unofficial": "http://www.inflacionverdadera.com/Argentina_inflation.xls",
                                "bra": f"https://servicodados.ibge.gov.br/api/v1/conjunturais?&d=s&user=ibge&t=1737&v=63&p=197001-{dt.datetime.now().strftime('%Y%m')}&ng=1(1)&c="},
                         "source": {"direct": [],
                                    "indirect": [
                                    "http://www.bcra.gov.ar/PublicacionesEstadisticas/Principales_variables_datos.asp",
                                    "http://www.inflacionverdadera.com/",
                                    "https://www.ibge.gov.br/estatisticas/economicas/precos-e-custos/9256-indice-nacional-de-precos-ao-consumidor-amplo.html?=&t=o-que-e"],
                                    "provider": ["econuy en base a BRCA, Inflación Verdadera e IBGE"]}},
        "regional_nxr": {"dl": {
            "ar": f"https://mercados.ambito.com/dolar/oficial/historico-general/09-04-2002/{dt.datetime.now().strftime('%d-%m-%Y')}",
            "ar_unofficial": f"https://mercados.ambito.com/dolar/informal/historico-general/09-04-2002/{dt.datetime.now().strftime('%d-%m-%Y')}",
            "bra": "http://www.ipeadata.gov.br/api/odata4/ValoresSerie(SERCODIGO='GM366_ERV366')"},
            "source": {"direct": [],
                       "indirect": ["https://www.ambito.com/contenidos/dolar-oficial.html",
                                    "https://www.ambito.com/contenidos/dolar-informal.html",
                                    "http://www.ipeadata.gov.br/Default.aspx"],
                       "provider": ["Ámbito", "Ipea"]}},
        "regional_policy_rates": {"dl": {
            "main": "https://www.bis.org/statistics/full_webstats_cbpol_d_dataflow_csv_row.zip"},
            "source": {"direct": [
                "https://www.bis.org/statistics/full_webstats_cbpol_d_dataflow_csv_row.zip"],
            "indirect": ["https://www.bis.org/statistics/cbpol.htm"],
            "provider": ["BIS"]}},
        "regional_monthly_gdp": {"dl": {
            "arg": "https://www.indec.gob.ar/ftp/cuadros/economia/sh_emae_mensual_base2004.xls",
            "bra": "http://api.bcb.gov.br/dados/serie/bcdata.sgs.24364/dados?formato=csv"},
            "source": {"direct": [],
                       "indirect": ["https://www.indec.gob.ar/indec/web/Nivel4-Tema-3-9-48",
                                    "https://dadosabertos.bcb.gov.br/dataset/24364-indice-de-atividade-economica-do-banco-central-ibc-br---com-ajuste-sazonal"],
                       "provider": ["INDEC", "BCB"]}},
        "regional_stocks": {"dl": {
            "arg": "https://www.investing.com/instruments/HistoricalDataAjax",
            "bra": f"https://query1.finance.yahoo.com/v7/finance/download/%5EBVSP?period1=735868800&period2={dt.datetime.now().timestamp().__round__()}&interval=1d&events=history&includeAdjustedClose=true"},
            "source": {"direct": [],
                       "indirect": [
                       "https://www.investing.com/indices/merv-historical-data",
                       "https://finance.yahoo.com/quote/%5EBVSP/history?period1=735868800&period2=1607212800&interval=1d&filter=history&frequency=1d&includeAdjustedClose=true"],
                       "provider": ["Investing.com", "Yahoo Finance"]}},
        "regional_rxr": {"dl": {},
                         "source": {"direct": [],
                                    "indirect": ["https://data.imf.org/?sk=4c514d48-b6ba-49ed-8ab9-52b0c1a0179b",
                                                 "https://www.ambito.com/contenidos/dolar-informal.html",
                                                 "http://www.bcra.gov.ar/PublicacionesEstadisticas/Principales_variables_datos.asp?serie=7931&detalle=Inflaci%F3n%20mensual%A0(variaci%F3n%20en%20%)",
                                                 "http://www.inflacionverdadera.com/argentina/",
                                                 "http://www.ipeadata.gov.br/Default.aspx",
                                                 "https://www.ibge.gov.br/estatisticas/economicas/precos-e-custos/9256-indice-nacional-de-precos-ao-consumidor-amplo.html?=&t=o-que-e"],
                                    "provider": ["econuy en base a BCRA, Ámbito, Ipea, IBGE, FMI e Inflación Verdadera"]}},
        }



import platform
import warnings
from datetime import datetime
from os import PathLike, getcwd, path
from pathlib import Path
from typing import Union, Optional, Tuple, Dict

import numpy as np
import pandas as pd
from statsmodels.tools.sm_exceptions import X13Error, X13Warning
from statsmodels.tsa import x13
from statsmodels.tsa.x13 import x13_arima_analysis as x13a
from statsmodels.tsa.seasonal import STL, seasonal_decompose

from econuy.utils import metadata



def convert_usd(df: pd.DataFrame,
                pipeline = None,
                errors: str = "raise") -> pd.DataFrame:
    """Convert to other units.

    See Also
    --------
    :mod:`~econuy.core.Pipeline.convert`.

    """
    if errors not in ["raise", "coerce", "ignore"]:
        raise ValueError("'errors' must be one of 'raise', "
                         "'coerce' or 'ignore'.")
    if "Moneda" not in df.columns.names:
        raise ValueError("Input dataframe's multiindex requires the "
                         "'Moneda' level.")

    if pipeline is None:
        from econuy.core import Pipeline
        pipeline = Pipeline()

    checks = [x == "UYU" for x in df.columns.get_level_values("Moneda")]
    if any(checks):
        if not all(checks) and errors == "raise":
            error_df = df.loc[:, [not check for check in checks]]
            msg = (f"{error_df.columns[0][0]} does not have the "
                   f"appropiate metadata.")
            return error_handler(df=df, errors=errors, msg=msg)
        pipeline.get(name="nxr_monthly")
        nxr_data = pipeline.dataset
        all_metadata = df.columns.droplevel("Indicador")
        if all(x == all_metadata[0] for x in all_metadata):
            return _convert_usd(df=df, nxr=nxr_data)
        else:
            columns = []
            for column_name, check in zip(df.columns, checks):
                df_column = df[[column_name]]
                if check is False:
                    msg = (f"{column_name[0]} does not have the "
                           f"appropiate metadata.")
                    columns.append(error_handler(df=df_column, errors=errors,
                                                 msg=msg))
                else:
                    converted = _convert_usd(df=df_column, nxr=nxr_data)
                    columns.append(converted)
            return pd.concat(columns, axis=1)
    else:
        return error_handler(df=df, errors=errors)


def _convert_usd(df: pd.DataFrame,
                 nxr: Optional[pd.DataFrame] = None) -> pd.DataFrame:
    if nxr is None:
        from econuy.core import Pipeline
        pipeline = Pipeline()
        pipeline.get("nxr_monthly")
        nxr = pipeline.dataset

    inferred_freq = pd.infer_freq(df.index)
    if inferred_freq in ["D", "B", "C", "W", "W-SUN", None]:
        if df.columns.get_level_values("Tipo")[0] == "Flujo":
            df = df.resample("M").sum()
        else:
            df = df.resample("M").last()
        inferred_freq = pd.infer_freq(df.index)

    if df.columns.get_level_values("Tipo")[0] == "Stock":
        metadata._set(nxr, ts_type="Stock")
        nxr_freq = resample(nxr, rule=inferred_freq,
                            operation="last").iloc[:, [1]]
    else:
        metadata._set(nxr, ts_type="Flujo")
        nxr_freq = resample(nxr, rule=inferred_freq,
                            operation="mean").iloc[:, [0]]
        cum_periods = int(df.columns.get_level_values("Acum. períodos")[0])
        nxr_freq = rolling(nxr_freq, window=cum_periods,
                           operation="mean")

    nxr_to_use = nxr_freq.reindex(df.index).iloc[:, 0]
    converted_df = df.div(nxr_to_use, axis=0)
    metadata._set(converted_df, currency="USD")

    return converted_df


def convert_real(df: pd.DataFrame,
                 start_date: Union[str, datetime, None] = None,
                 end_date: Union[str, datetime, None] = None,
                 pipeline = None,
                 errors: str = "raise") -> pd.DataFrame:
    """Convert to other units.

    See Also
    --------
    :mod:`~econuy.core.Pipeline.convert`.

    """
    if errors not in ["raise", "coerce", "ignore"]:
        raise ValueError("'errors' must be one of 'raise', "
                         "'coerce' or 'ignore'.")
    if "Inf. adj." not in df.columns.names:
        raise ValueError("Input dataframe's multiindex requires the "
                         "'Inf. adj.' level.")

    if pipeline is None:
        from econuy.core import Pipeline
        pipeline = Pipeline()

    checks = [x == "UYU" and "Const." not in y
              for x, y in zip(df.columns.get_level_values("Moneda"),
                              df.columns.get_level_values("Inf. adj."))]
    if any(checks):
        if not all(checks) and errors == "raise":
            error_df = df.loc[:, [not check for check in checks]]
            msg = (f"{error_df.columns[0][0]} does not have the "
                   f"appropiate metadata.")
            return error_handler(df=df, errors=errors, msg=msg)
        pipeline.get(name="cpi")
        cpi_data = pipeline.dataset
        all_metadata = df.columns.droplevel("Indicador")
        if all(x == all_metadata[0] for x in all_metadata):
            return _convert_real(df=df, start_date=start_date,
                                 end_date=end_date, cpi=cpi_data)
        else:
            columns = []
            for column_name, check in zip(df.columns, checks):
                df_column = df[[column_name]]
                if check is False:
                    msg = (f"{column_name[0]} does not have the "
                           f"appropiate metadata.")
                    columns.append(error_handler(df=df_column, errors=errors,
                                                 msg=msg))
                else:
                    converted = _convert_real(df=df_column,
                                              start_date=start_date,
                                              end_date=end_date, cpi=cpi_data)
                    columns.append(converted)
            return pd.concat(columns, axis=1)
    else:
        return error_handler(df=df, errors=errors)


def _convert_real(df: pd.DataFrame,
                  start_date: Union[str, datetime, None] = None,
                  end_date: Union[str, datetime, None] = None,
                  cpi: Optional[pd.DataFrame] = None) -> pd.DataFrame:
    if cpi is None:
        from econuy.core import Pipeline
        pipeline = Pipeline()
        pipeline.get("cpi")
        cpi = pipeline.dataset

    inferred_freq = pd.infer_freq(df.index)
    if inferred_freq in ["D", "B", "C", "W", "W-SUN", None]:
        if df.columns.get_level_values("Tipo")[0] == "Flujo":
            df = df.resample("M").sum()
        else:
            df = df.resample("M").mean()
        inferred_freq = pd.infer_freq(df.index)

    metadata._set(cpi, ts_type="Flujo")
    cpi_freq = resample(cpi, rule=inferred_freq,
                        operation="mean").iloc[:, [0]]
    cum_periods = int(df.columns.get_level_values("Acum. períodos")[0])
    cpi_to_use = rolling(cpi_freq, window=cum_periods,
                         operation="mean").squeeze()

    if start_date is None:
        converted_df = df.div(cpi_to_use, axis=0)
        col_text = "Const."
    elif end_date is None:
        month = df.iloc[df.index.get_loc(start_date, method="nearest")].name
        converted_df = df.div(cpi_to_use, axis=0) * cpi_to_use.loc[month]
        m_start = datetime.strptime(start_date, "%Y-%m-%d").strftime("%Y-%m")
        col_text = f"Const. {m_start}"
    else:
        converted_df = df.div(cpi_to_use, axis=0) * cpi_to_use[start_date:
                                                               end_date].mean()
        m_start = datetime.strptime(start_date, "%Y-%m-%d").strftime("%Y-%m")
        m_end = datetime.strptime(end_date, "%Y-%m-%d").strftime("%Y-%m")
        if m_start == m_end:
            col_text = f"Const. {m_start}"
        else:
            col_text = f"Const. {m_start}_{m_end}"

    converted_df = converted_df.reindex(df.index)
    metadata._set(converted_df, inf_adj=col_text)

    return converted_df


def convert_gdp(df: pd.DataFrame,
                pipeline = None,
                errors: str = "raise") -> pd.DataFrame:
    """Convert to other units.

    See Also
    --------
    :mod:`~econuy.core.Pipeline.convert`.

    """
    if errors not in ["raise", "coerce", "ignore"]:
        raise ValueError("'errors' must be one of 'raise', 'coerce' or "
                         "'ignore'.")
    if any(x not in df.columns.names for x in ["Área", "Unidad"]):
        raise ValueError("Input dataframe's multiindex requires the 'Área' "
                         "and 'Unidad' levels.")

    if pipeline is None:
        from econuy.core import Pipeline
        pipeline = Pipeline()

    checks = [x not in ["Regional", "Global"] and "%PBI" not in y
              for x, y in zip(df.columns.get_level_values("Área"),
                              df.columns.get_level_values("Unidad"))]
    if any(checks):
        if not all(checks) and errors == "raise":
            error_df = df.loc[:, [not check for check in checks]]
            msg = (f"{error_df.columns[0][0]} does not have the "
                   f"appropiate metadata.")
            return error_handler(df=df, errors=errors, msg=msg)
        pipeline.get(name="_lin_gdp")
        gdp_data = pipeline.dataset
        all_metadata = df.columns.droplevel("Indicador")
        if all(x == all_metadata[0] for x in all_metadata):
            return _convert_gdp(df=df, gdp=gdp_data)
        else:
            columns = []
            for column_name, check in zip(df.columns, checks):
                df_column = df[[column_name]]
                if check is False:
                    msg = (f"{column_name[0]} does not have the "
                           f"appropiate metadata.")
                    columns.append(error_handler(df=df_column, errors=errors,
                                                 msg=msg))
                else:
                    converted = _convert_gdp(df=df_column, gdp=gdp_data)
                    columns.append(converted)
            return pd.concat(columns, axis=1)
    else:
        return error_handler(df=df, errors=errors)


def _convert_gdp(df: pd.DataFrame,
                 gdp: Optional[pd.DataFrame] = None) -> pd.DataFrame:
    if gdp is None:
        from econuy.core import Pipeline
        pipeline = Pipeline()
        pipeline.get("_lin_gdp")
        gdp = pipeline.dataset

    inferred_freq = pd.infer_freq(df.index)
    cum = int(df.columns.get_level_values("Acum. períodos")[0])
    if inferred_freq in ["M", "MS"]:
        gdp = resample(gdp, rule=inferred_freq,
                       operation="upsample", interpolation="linear")
        if cum != 12 and df.columns.get_level_values("Tipo")[0] == "Flujo":
            converter = int(12 / cum)
            df = rolling(df, window=converter, operation="sum")
    elif inferred_freq in ["Q", "Q-DEC"]:
        gdp = gdp.resample(inferred_freq, convention="end").asfreq()
        if cum != 4 and df.columns.get_level_values("Tipo")[0] == "Flujo":
            converter = int(4 / cum)
            df = rolling(df, window=converter, operation="sum")
    elif inferred_freq in ["A", "A-DEC"]:
        gdp = gdp.resample(inferred_freq, convention="end").asfreq()
    elif inferred_freq in ["D", "B", "C", "W", "W-SUN", None]:
        if df.columns.get_level_values("Tipo")[0] == "Flujo":
            df = df.resample("M").sum()
        else:
            df = df.resample("M").mean()
        gdp = resample(gdp, rule="M",
                       operation="upsample", interpolation="linear")
    else:
        raise ValueError("Frequency of input dataframe not any of 'D', 'C', "
                         "'W', 'B', 'M', 'MS', 'Q', 'Q-DEC', 'A' or 'A-DEC'.")

    if df.columns.get_level_values("Moneda")[0] == "USD":
        gdp = gdp.iloc[:, 1].to_frame()
    else:
        gdp = gdp.iloc[:, 0].to_frame()

    gdp_to_use = gdp.reindex(df.index).iloc[:, 0]
    converted_df = df.div(gdp_to_use, axis=0).multiply(100)

    metadata._set(converted_df, unit="% PBI")

    return converted_df


def resample(df: pd.DataFrame, rule: Union[pd.DateOffset, pd.Timedelta, str],
             operation: str = "sum",
             interpolation: str = "linear",
             warn: bool = False) -> pd.DataFrame:
    """
    Resample to target frequencies.

    See Also
    --------
    :mod:`~econuy.core.Pipeline.resample`

    """
    if operation not in ["sum", "mean", "upsample", "last"]:
        raise ValueError("Invalid 'operation' option.")
    if "Acum. períodos" not in df.columns.names:
        raise ValueError("Input dataframe's multiindex requires the "
                         "'Acum. períodos' level.")

    all_metadata = df.columns.droplevel("Indicador")
    if all(x == all_metadata[0] for x in all_metadata):
        return _resample(df=df, rule=rule, operation=operation,
                         interpolation=interpolation, warn=warn)
    else:
        columns = []
        for column_name in df.columns:
            df_column = df[[column_name]]
            converted = _resample(df=df_column, rule=rule, operation=operation,
                                  interpolation=interpolation, warn=warn)
            columns.append(converted)
        return pd.concat(columns, axis=1)


def _resample(df: pd.DataFrame, rule: Union[pd.DateOffset, pd.Timedelta, str],
              operation: str = "sum",
              interpolation: str = "linear",
              warn: bool = False) -> pd.DataFrame:
    pd_frequencies = {"A": 1,
                      "A-DEC": 1,
                      "Q": 4,
                      "Q-DEC": 4,
                      "M": 12,
                      "W": 52.143,
                      "W-SUN": 52.143,
                      "2W": 26.071,
                      "2W-SUN": 26.071,
                      "B": 240,
                      "D": 365}

    if operation == "sum":
        resampled_df = df.resample(rule).sum()
    elif operation == "mean":
        resampled_df = df.resample(rule).mean()
    elif operation == "last":
        resampled_df = df.resample(rule).last()
    else:
        resampled_df = df.resample(rule).last()
        resampled_df = resampled_df.interpolate(method=interpolation)

    cum_periods = int(df.columns.get_level_values("Acum. períodos")[0])
    if cum_periods != 1:
        input_notna = df.iloc[:, 0].count()
        output_notna = resampled_df.iloc[:, 0].count()
        cum_adj = round(output_notna / input_notna)
        metadata._set(resampled_df,
                      cumperiods=int(cum_periods * cum_adj))

    if operation in ["sum", "mean", "last"]:
        infer_base = pd.infer_freq(df.index)
        try:
            base_freq = pd_frequencies[infer_base]
            target_freq = pd_frequencies[rule]
            if target_freq < base_freq:
                count = int(base_freq / target_freq)
                proc = df.resample(rule).count()
                antimask = np.where(proc >= count, False, True)
                resampled_df = resampled_df.mask(antimask, np.nan)
        except KeyError:
            if warn:
                warnings.warn("No bin trimming performed because frequencies "
                              "could not be assigned a numeric value",
                              UserWarning)

    metadata._set(resampled_df)
    resampled_df = resampled_df.dropna(how="all")

    return resampled_df


def rolling(df: pd.DataFrame, window: Optional[int] = None,
            operation: str = "sum") -> pd.DataFrame:
    """
    Calculate rolling averages or sums.

    See Also
    --------
    :mod:`~econuy.core.Pipeline.rolling`.

    """
    if operation not in ["sum", "mean"]:
        raise ValueError("Invalid 'operation' option.")
    if "Tipo" not in df.columns.names:
        raise ValueError("Input dataframe's multiindex requires the "
                         "'Tipo' level.")

    all_metadata = df.columns.droplevel("Indicador")
    if all(x == all_metadata[0] for x in all_metadata):
        return _rolling(df=df, window=window, operation=operation)
    else:
        columns = []
        for column_name in df.columns:
            df_column = df[[column_name]]
            converted = _rolling(df=df_column, window=window,
                                 operation=operation)
            columns.append(converted)
        return pd.concat(columns, axis=1)


def _rolling(df: pd.DataFrame, window: Optional[int] = None,
             operation: str = "sum") -> pd.DataFrame:
    pd_frequencies = {"A": 1,
                      "A-DEC": 1,
                      "Q": 4,
                      "Q-DEC": 4,
                      "M": 12,
                      "MS": 12,
                      "W": 52,
                      "W-SUN": 52,
                      "2W": 26,
                      "2W-SUN": 26,
                      "B": 260,
                      "D": 365}

    window_operation = {"sum": lambda x: x.rolling(window=window,
                                                   min_periods=window).sum(),
                        "mean": lambda x: x.rolling(window=window,
                                                    min_periods=window).mean()}

    if df.columns.get_level_values("Tipo")[0] == "Stock":
        warnings.warn("Rolling operations should not be "
                      "calculated on stock variables", UserWarning)

    if window is None:
        inferred_freq = pd.infer_freq(df.index)
        window = pd_frequencies[inferred_freq]

    rolling_df = df.apply(window_operation[operation])

    metadata._set(rolling_df, cumperiods=window)

    return rolling_df


def rebase(df: pd.DataFrame, start_date: Union[str, datetime],
           end_date: Union[str, datetime, None] = None,
           base: Union[int, float] = 100.0) -> pd.DataFrame:
    """
    Scale to a period or range of periods.

    See Also
    --------
    :mod:`~econuy.core.Pipeline.rebase`.

    """
    all_metadata = df.columns.droplevel("Indicador")
    if all(x == all_metadata[0] for x in all_metadata):
        return _rebase(df=df, end_date=end_date,
                       start_date=start_date, base=base)
    else:
        columns = []
        for column_name in df.columns:
            df_column = df[[column_name]]
            converted = _rebase(df=df_column, end_date=end_date,
                                start_date=start_date, base=base)
            columns.append(converted)
        return pd.concat(columns, axis=1)


def _rebase(df: pd.DataFrame, start_date: Union[str, datetime],
            end_date: Union[str, datetime, None] = None,
            base: float = 100.0) -> pd.DataFrame:
    if end_date is None:
        start_date = df.iloc[df.index.get_loc(start_date,
                                              method="nearest")].name
        indexed = df.apply(lambda x: x / x.loc[start_date] * base)
        if isinstance(start_date, str):
            start_date = datetime.strptime(start_date, "%Y-%m-%d")
        if not isinstance(base, int):
            if base.is_integer():
                base = int(base)
        m_start = start_date.strftime("%Y-%m")
        metadata._set(indexed, unit=f"{m_start}={base}")

    else:
        indexed = df.apply(lambda x: x / x[start_date:end_date].mean() * base)
        if isinstance(start_date, str):
            start_date = datetime.strptime(start_date, "%Y-%m-%d")
        if isinstance(end_date, str):
            end_date = datetime.strptime(end_date, "%Y-%m-%d")
        m_start = start_date.strftime("%Y-%m")
        m_end = end_date.strftime("%Y-%m")
        if not isinstance(base, int):
            if base.is_integer():
                base = int(base)
        if m_start == m_end:
            metadata._set(indexed, unit=f"{m_start}={base}")
        else:
            metadata._set(indexed, unit=f"{m_start}_{m_end}={base}")

    return indexed


# The `_open_and_read` function needs to be monkey-patched to specify the
# encoding or decomposition will fail on Windows
def _new_open_and_read(fname):
    with open(fname, 'r', encoding='utf8') as fin:
        fout = fin.read()
    return fout


x13._open_and_read = _new_open_and_read


def decompose(df: pd.DataFrame, component: str = "both", method: str = "x13",
              force_x13: bool = False, fallback: str = "loess",
              outlier: bool = True, trading: bool = True,
              x13_binary: Union[str, PathLike, None] = "search",
              search_parents: int = 1, ignore_warnings: bool = True,
              errors: str = "raise",
              **kwargs) -> Union[Dict[str, pd.DataFrame],
                                 pd.DataFrame]:
    """
    Apply seasonal decomposition.

    By default returns both trend and seasonally adjusted components,
    unlike the class method referred below.

    See Also
    --------
    :mod:`~econuy.core.Pipeline.decompose`.

    """
    if errors not in ["raise", "coerce", "ignore"]:
        raise ValueError("method can only be 'x13', 'loess' or 'ma'.")
    if method not in ["x13", "loess", "ma"]:
        raise ValueError("method can only be 'x13', 'loess' or 'ma'.")
    if fallback not in ["loess", "ma"]:
        raise ValueError("method can only be 'loess' or 'ma'.")
    if component not in ["trend", "seas", "both"]:
        raise ValueError("component can only be 'trend', 'seas' or 'both'.")
    if "Seas. Adj." not in df.columns.names:
        raise ValueError("Input dataframe's multiindex requires the "
                         "'Seas. Adj.' level.")

    binary_path = None
    if method == "x13":
        if x13_binary == "search":
            search_term = "x13as"
            if platform.system() == "Windows":
                search_term += ".exe"
            binary_path = _rsearch(dir_file=getcwd(), n=search_parents,
                                   search_term=search_term)
        elif isinstance(x13_binary, str):
            binary_path = x13_binary
        elif isinstance(x13_binary, PathLike):
            binary_path = Path(x13_binary).as_posix()
        else:
            binary_path = None
        if isinstance(binary_path, str) and path.isfile(
                binary_path) is False:
            raise FileNotFoundError(
                "X13 binary missing. Please refer to the README "
                "for instructions on where to get binaries for "
                "Windows and Unix, and how to compile it for "
                "macOS.")

    checks = [x not in ["Tendencia", "SA"]
              for x in df.columns.get_level_values("Seas. Adj.")]
    passing = df.loc[:, checks]
    not_passing = df.loc[:, [not x for x in checks]]
    if any(checks):
        if not all(checks) and errors == "raise":
            error_df = df.loc[:, [not check for check in checks]]
            msg = (f"{error_df.columns[0][0]} does not have the "
                   f"appropiate metadata.")
            return error_handler(df=df, errors=errors, msg=msg)
        passing_output = _decompose(passing, component=component,
                                    method=method, force_x13=force_x13,
                                    fallback=fallback, outlier=outlier,
                                    trading=trading, x13_binary=binary_path,
                                    ignore_warnings=ignore_warnings,
                                    errors=errors, **kwargs)
        if not_passing.shape[1] != 0:
            not_passing_output = error_handler(df=not_passing, errors=errors)
        else:
            not_passing_output = not_passing
        if isinstance(passing_output, pd.DataFrame):
            output = pd.concat([passing_output, not_passing_output], axis=1)
            output = output[df.columns.get_level_values(0)]
            return output
        elif isinstance(passing_output, Dict):
            output = {}
            for name, data in passing_output.items():
                aux = pd.concat([data, not_passing_output], axis=1)
                output[name] = aux[df.columns.get_level_values(0)]
            return output
    else:
        return error_handler(df=df, errors=errors)


def _decompose(df: pd.DataFrame, component: str = "both", method: str = "x13",
               force_x13: bool = False, fallback: str = "loess",
               outlier: bool = True, trading: bool = True,
               x13_binary: Union[str, PathLike, None] = None,
               ignore_warnings: bool = True, errors: str = "raise",
               **kwargs) -> Union[Tuple[pd.DataFrame, pd.DataFrame],
                                  pd.DataFrame]:
    if method not in ["x13", "loess", "ma"]:
        raise ValueError("method can only be 'x13', 'loess' or 'ma'.")
    if fallback not in ["loess", "ma"]:
        raise ValueError("method can only be 'loess' or 'ma'.")

    df_proc = df.copy()
    old_columns = df_proc.columns
    df_proc.columns = df_proc.columns.get_level_values(level=0)
    df_proc.index = pd.to_datetime(df_proc.index, errors="coerce")
    trends = pd.DataFrame(data=np.nan, index=df_proc.index,
                          columns=old_columns)
    seas_adjs = trends.copy()

    if method == "x13":
        try:
            with warnings.catch_warnings():
                if ignore_warnings is True:
                    action = "ignore"
                else:
                    action = "default"
                warnings.filterwarnings(action=action, category=X13Warning)
                results = df_proc.apply(
                    lambda x: x13a(x.dropna(), outlier=outlier,
                                   trading=trading, x12path=x13_binary,
                                   prefer_x13=True, **kwargs)
                )
            trends = results.apply(lambda x: x.trend.reindex(df_proc.index)).T
            seas_adjs = results.apply(lambda x: x.seasadj.
                                      reindex(df_proc.index)).T

        except X13Error:
            if force_x13 is True:
                if outlier is True:
                    try:
                        warnings.warn("X13 error found with selected "
                                      "parameters. Trying with outlier=False.",
                                      UserWarning)
                        return decompose(df=df, method=method,
                                         outlier=False,
                                         component=component,
                                         fallback=fallback,
                                         force_x13=force_x13,
                                         x13_binary=x13_binary,
                                         **kwargs)
                    except X13Error:
                        try:
                            warnings.warn("X13 error found with trading=True. "
                                          "Trying with trading=False.",
                                          UserWarning)
                            return decompose(df=df, method=method,
                                             outlier=False, trading=False,
                                             component=component,
                                             fallback=fallback,
                                             force_x13=force_x13,
                                             x13_binary=x13_binary,
                                             **kwargs)
                        except X13Error:
                            warnings.warn("No combination of parameters "
                                          "successful. No decomposition "
                                          "performed.",
                                          UserWarning)
                            trends = error_handler(df=df_proc, errors=errors)
                            seas_adjs = trends.copy()

                elif trading is True:
                    try:
                        warnings.warn("X13 error found with trading=True. "
                                      "Trying with trading=False...",
                                      UserWarning)
                        return decompose(df=df, method=method,
                                         trading=False, component=component,
                                         fallback=fallback,
                                         force_x13=force_x13,
                                         x13_binary=x13_binary,
                                         **kwargs)
                    except X13Error:
                        warnings.warn("No combination of parameters "
                                      "successful. Filling with NaN.",
                                      UserWarning)
                        trends = error_handler(df=df_proc, errors=errors)
                        seas_adjs = trends.copy()

            else:
                if fallback == "loess":
                    results = df_proc.apply(
                        lambda x: STL(x.dropna()).fit(), result_type="expand")
                else:
                    results = df_proc.apply(
                        lambda x: seasonal_decompose(
                            x.dropna(), extrapolate_trend="freq"),
                        result_type="expand")
                trends = results.apply(lambda x:
                                       x.trend.reindex(df_proc.index)).T
                seas_adjs = results.apply(
                    lambda x: (x.observed
                               - x.seasonal).reindex(df_proc.index)).T

    else:
        if method == "loess":
            results = df_proc.apply(
                lambda x: STL(x.dropna()).fit(), result_type="expand")
        else:
            results = df_proc.apply(
                lambda x: seasonal_decompose(x.dropna(),
                                             extrapolate_trend="freq"),
                result_type="expand")
        trends = results.apply(lambda x:
                               x.trend.reindex(df_proc.index)).T
        seas_adjs = results.apply(
            lambda x: (x.observed - x.seasonal).reindex(df_proc.index)).T

    trends.columns = old_columns
    seas_adjs.columns = old_columns
    metadata._set(trends, seas_adj="Tendencia")
    metadata._set(seas_adjs, seas_adj="SA")
    output = pd.DataFrame()
    if component == "both":
        output = {"trend": trends, "seas": seas_adjs}
    elif component == "seas":
        output = seas_adjs
    elif component == "trend":
        output = trends

    return output


def _rsearch(dir_file: Union[str, PathLike], search_term: str, n: int = 2):
    """Recursively search for a file starting from the n-parent folder of
    a supplied path."""
    i = 0
    while i < n:
        i += 1
        dir_file = path.dirname(dir_file)
    try:
        final_path = ([x for x in Path(dir_file).rglob(search_term)][0]
                      .absolute().as_posix())
    except IndexError:
        final_path = True
    return final_path


def chg_diff(df: pd.DataFrame, operation: str = "chg",
             period: str = "last") -> pd.DataFrame:
    """
    Calculate pct change or difference.

    See Also
    --------
    :mod:`~econuy.core.Pipeline.chg_diff`.

    """
    if operation not in ["chg", "diff"]:
        raise ValueError("Invalid 'operation' option.")
    if period not in ["last", "inter", "annual"]:
        raise ValueError("Invalid 'period' option.")
    if "Tipo" not in df.columns.names:
        raise ValueError("Input dataframe's multiindex requires the "
                         "'Tipo' level.")

    all_metadata = df.columns.droplevel("Indicador")
    if all(x == all_metadata[0] for x in all_metadata):
        return _chg_diff(df=df, operation=operation, period=period)
    else:
        columns = []
        for column_name in df.columns:
            df_column = df[[column_name]]
            converted = _chg_diff(df=df_column, operation=operation,
                                  period=period)
            columns.append(converted)
        return pd.concat(columns, axis=1)


def _chg_diff(df: pd.DataFrame, operation: str = "chg",
              period: str = "last") -> pd.DataFrame:
    inferred_freq = pd.infer_freq(df.index)

    type_change = {"last":
                   {"chg": [lambda x: x.pct_change(periods=1),
                            "% variación"],
                    "diff": [lambda x: x.diff(periods=1), "Cambio"]},
                   "inter":
                       {"chg": [lambda x: x.pct_change(periods=last_year),
                                "% variación interanual"],
                        "diff": [lambda x: x.diff(periods=last_year),
                                 "Cambio interanual"]},
                   "annual":
                       {"chg": [lambda x: x.pct_change(periods=last_year),
                                "% variación anual"],
                        "diff": [lambda x: x.diff(periods=last_year),
                                 "Cambio anual"]}}

    if inferred_freq == "M":
        last_year = 12
    elif inferred_freq == "Q" or inferred_freq == "Q-DEC":
        last_year = 4
    elif inferred_freq == "A" or inferred_freq == "A-DEC":
        last_year = 1
    else:
        raise ValueError("The dataframe needs to have a frequency of M "
                         "(month end), Q (quarter end) or A (year end)")

    if period == "annual":

        if df.columns.get_level_values("Tipo")[0] == "Stock":
            output = df.apply(type_change[period][operation][0])
        else:
            output = rolling(df, operation="sum")
            output = output.apply(
                type_change[period][operation][0])

        metadata._set(output, unit=type_change[period][operation][1])

    else:
        output = df.apply(type_change[period][operation][0])
        metadata._set(output, unit=type_change[period][operation][1])

    if operation == "chg":
        output = output.multiply(100)

    return output


def error_handler(df: pd.DataFrame, errors: str,
                  msg: str = None) -> pd.DataFrame:
    if errors == "coerce":
        return pd.DataFrame(data=np.nan, index=df.index, columns=df.columns)
    elif errors == "ignore":
        return df
    elif errors == "raise":
        if msg is None:
            msg = ""
        raise ValueError(msg)



import re
import tempfile
import time
import zipfile
import datetime as dt
from random import randint
from io import BytesIO
from os import path, listdir
from typing import  Optional
from urllib.error import HTTPError, URLError

import pandas as pd
import numpy as np
import requests
from pandas.tseries.offsets import MonthEnd
from bs4 import BeautifulSoup
from opnieuw import retry
from selenium.webdriver.remote.webdriver import WebDriver

from econuy.core import Pipeline
from econuy.transform import rebase, resample
from econuy.utils import metadata
from econuy.utils.chromedriver import _build
from econuy.utils.sources import urls
from econuy.utils.extras import investing_headers


@retry(
    retry_on_exceptions=(HTTPError, URLError),
    max_calls_total=12,
    retry_window_after_first_call_in_seconds=60,
)
def gdp(driver: WebDriver = None) -> pd.DataFrame:
    """Get seasonally adjusted real GDP for Argentina and Brazil.

    This function requires a Selenium webdriver. It can be provided in the
    driver parameter, or it will attempt to configure a Chrome webdriver.

    Parameters
    ----------
    driver : selenium.webdriver.chrome.webdriver.WebDriver, default None
        Selenium webdriver for scraping. If None, build a Chrome webdriver.

    Returns
    -------
    Quarterly real GDP : pd.DataFrame

    """
    name = "regional_gdp"

    if driver is None:
        driver = _build()
    driver.get(urls[name]["dl"]["arg_new"])
    time.sleep(5)
    soup = BeautifulSoup(driver.page_source, "lxml")
    driver.quit()
    url = soup.find_all(href=re.compile("desest"))[0]["href"]
    full_url = f"https://www.indec.gob.ar{url}"
    arg = pd.read_excel(full_url, skiprows=3,
                        usecols="D").dropna(how="all")
    arg.index = pd.date_range(start="2004-03-31", freq="Q-DEC",
                              periods=len(arg))
    arg_old = pd.read_excel(urls[name]["dl"]["arg_old"], skiprows=7,
                            usecols="D").dropna(how="all")
    arg_old.index = pd.date_range(start="1993-03-31", freq="Q-DEC",
                                  periods=len(arg_old))
    arg = pd.concat([arg, arg_old], axis=1)
    for row in reversed(range(len(arg))):
        if pd.isna(arg.iloc[row, 0]):
            arg.iloc[row, 0] = (arg.iloc[row, 1] / arg.iloc[row + 1, 1]
                                * arg.iloc[row + 1, 0])
    arg = arg.iloc[:, [0]]

    r = requests.get(urls[name]["dl"]["bra"])
    temp_dir = tempfile.TemporaryDirectory()
    with zipfile.ZipFile(BytesIO(r.content), "r") as f:
        f.extractall(path=temp_dir.name)
    path_temp = path.join(temp_dir.name, listdir(temp_dir.name)[0])
    bra = pd.read_excel(path_temp, usecols="Q", skiprows=3,
                        sheet_name="Val encad preços 95 com ajuste")
    bra.index = pd.date_range(start="1996-03-31", freq="Q-DEC",
                              periods=len(bra))

    output = pd.concat([arg, bra], axis=1).div(1000)
    output.columns = ["Argentina", "Brasil"]
    output.rename_axis(None, inplace=True)

    metadata._set(output, area="Regional", currency="-",
                  inf_adj="Const.", seas_adj="SA", unit="Miles de millones",
                  ts_type="Flujo", cumperiods=1)
    metadata._modify_multiindex(output, levels=[3],
                                new_arrays=[["ARS", "BRL"]])

    return output


@retry(
    retry_on_exceptions=(HTTPError, URLError),
    max_calls_total=4,
    retry_window_after_first_call_in_seconds=60,
)
def monthly_gdp() -> pd.DataFrame:
    """Get monthly GDP data.

    Countries/aggregates selected are Argentina and Brazil.

    Returns
    -------
    Daily policy interest rates : pd.DataFrame

    """
    name = "regional_monthly_gdp"

    arg = pd.read_excel(urls[name]["dl"]["arg"], usecols="D",
                        skiprows=4).dropna(how="all")
    arg.index = pd.date_range(start="2004-01-31", freq="M", periods=len(arg))

    bra = pd.read_csv(urls[name]["dl"]["bra"], sep=";",
                      index_col=0, decimal=",")
    bra.index = pd.date_range(start="2003-01-31", freq="M", periods=len(bra))

    output = pd.concat([arg, bra], axis=1)
    output.columns = ["Argentina", "Brasil"]
    output.rename_axis(None, inplace=True)
    metadata._set(output, area="Regional", currency="-",
                  inf_adj="Const.", seas_adj="SA",
                  ts_type="Flujo", cumperiods=1)
    metadata._modify_multiindex(output, levels=[3],
                                new_arrays=[["ARS", "BRL"]])
    output = rebase(output, start_date="2010-01-01", end_date="2010-12-31")

    return output


@retry(
    retry_on_exceptions=(HTTPError, URLError),
    max_calls_total=12,
    retry_window_after_first_call_in_seconds=60,
)
def cpi() -> pd.DataFrame:
    """Get consumer price index for Argentina and Brazil.

    Returns
    -------
    Monthly CPI : pd.DataFrame

    """
    name = "regional_cpi"

    arg = requests.get(urls[name]["dl"]["ar"],
                       params=urls[name]["dl"]["ar_payload"])
    arg = pd.read_html(arg.content)[0]
    arg.set_index("Fecha", drop=True, inplace=True)
    arg.index = pd.to_datetime(arg.index, format="%d/%m/%Y")
    arg.columns = ["nivel"]
    arg = arg.divide(10)

    arg_unoff = pd.read_excel(urls[name]["dl"]["ar_unofficial"])
    arg_unoff.set_index("date", drop=True, inplace=True)
    arg_unoff.index = arg_unoff.index + MonthEnd(0)
    arg_unoff = arg_unoff.loc[(arg_unoff.index >= "2006-12-01") &
                              (arg_unoff.index <= "2016-12-01"), "index"]
    arg_unoff = arg_unoff.to_frame().pct_change(periods=1).multiply(100).dropna()
    arg_unoff.columns = ["nivel"]
    arg = (arg.append(arg_unoff).reset_index().
           drop_duplicates(subset="index", keep="last").
           set_index("index", drop=True).sort_index())
    arg = arg.divide(100).add(1).cumprod()

    bra_r = requests.get(urls[name]["dl"]["bra"])
    bra = pd.DataFrame(bra_r.json())[["v"]]
    bra.index = pd.date_range(start="1979-12-31", freq="M", periods=len(bra))
    bra = bra.apply(pd.to_numeric, errors="coerce")
    bra = bra.divide(100).add(1).cumprod()

    output = pd.concat([arg, bra], axis=1)
    output.columns = ["Argentina", "Brasil"]
    output.rename_axis(None, inplace=True)
    metadata._set(output, area="Regional", currency="-",
                  inf_adj="No", seas_adj="NSA",
                  ts_type="-", cumperiods=1)
    metadata._modify_multiindex(output, levels=[3],
                                new_arrays=[["ARS", "BRL"]])
    output = rebase(output, start_date="2010-10-01", end_date="2010-10-31")

    return output


@retry(
    retry_on_exceptions=(HTTPError, URLError),
    max_calls_total=12,
    retry_window_after_first_call_in_seconds=60,
)
def embi_spreads() -> pd.DataFrame:
    """Get EMBI spread for Argentina, Brazil and the EMBI Global.

    Returns
    -------
    Daily 10-year government bond spreads : pd.DataFrame

    """
    name = "regional_embi_spreads"

    global_ = pd.read_excel(urls[name]["dl"]["global"],
                            usecols="A:B", skiprows=1, index_col=0,
                            parse_dates=True)
    global_ = global_.loc[~pd.isna(global_.index)].mul(100)
    region = []
    for cnt in ["argentina", "brasil"]:
        r = requests.get(urls[name]["dl"][cnt])
        aux = pd.DataFrame(r.json())
        aux.set_index(0, drop=True, inplace=True)
        aux.drop("Fecha", inplace=True)
        aux = aux.replace(",", ".", regex=True).apply(pd.to_numeric)
        aux.index = pd.to_datetime(aux.index, format="%d-%m-%Y")
        aux.sort_index(inplace=True)
        aux.columns = [cnt]
        region.append(aux)
    region = region[0].join(region[1]).interpolate(limit_area="inside")
    output = region.join(global_, how="left").interpolate(method="linear",
                                                          limit_area="inside")
    output.columns = ["Argentina", "Brasil", "EMBI Global"]
    output = output.apply(pd.to_numeric, errors="coerce")
    output.rename_axis(None, inplace=True)

    metadata._set(output, area="Regional", currency="USD",
                  inf_adj="No", seas_adj="NSA", unit="PBS",
                  ts_type="-", cumperiods=1)

    return output


@retry(
    retry_on_exceptions=(HTTPError, URLError),
    max_calls_total=12,
    retry_window_after_first_call_in_seconds=60,
)
def embi_yields(pipeline: Optional[Pipeline] = None) -> pd.DataFrame:
    """Get EMBI yields for Argentina, Brazil and the EMBI Global.

    Yields are calculated by adding EMBI spreads to the 10-year US Treasury
    bond rate.

    Parameters
    ----------
    pipeline : econuy.core.Pipeline or None, default None
        An instance of the econuy Pipeline class.

    Returns
    -------
    Daily 10-year government bonds interest rates : pd.DataFrame

    """
    name = "regional_embi_yields"

    if pipeline is None:
        pipeline = Pipeline()

    pipeline.get("global_long_rates")
    treasuries = pipeline.dataset["Estados Unidos"]
    pipeline.get("regional_embi_spreads")
    spreads = pipeline.dataset

    treasuries = (treasuries.reindex(spreads.index)
                  .interpolate(method="linear", limit_direction="forward"))
    output = spreads.div(100).add(treasuries.squeeze(), axis=0)
    output.rename_axis(None, inplace=True)

    metadata._set(output, area="Regional", currency="USD",
                  inf_adj="No", seas_adj="NSA", unit="Tasa",
                  ts_type="-", cumperiods=1)

    return output


@retry(
    retry_on_exceptions=(HTTPError, URLError),
    max_calls_total=12,
    retry_window_after_first_call_in_seconds=60,
)
def nxr() -> pd.DataFrame:
    """Get USDARS and USDBRL.

    Returns
    -------
    Daily exchange rates : pd.DataFrame

    """
    name = "regional_nxr"

    arg = []
    for dollar in ["ar", "ar_unofficial"]:
        r = requests.get(urls[name]["dl"][dollar])
        aux = pd.DataFrame(r.json())[[0, 2]]
        aux.set_index(0, drop=True, inplace=True)
        aux.drop("Fecha", inplace=True)
        aux = aux.replace(",", ".", regex=True).apply(pd.to_numeric)
        aux.index = pd.to_datetime(aux.index, format="%d-%m-%Y")
        aux.sort_index(inplace=True)
        aux.columns = [dollar]
        arg.append(aux)
    arg = arg[0].join(arg[1], how="left")
    arg.columns = ["Argentina - oficial", "Argentina - informal"]

    r = requests.get(urls[name]["dl"]["bra"])
    bra = pd.DataFrame(r.json())
    bra = [(x["VALDATA"], x["VALVALOR"]) for x in bra["value"]]
    bra = pd.DataFrame.from_records(bra).dropna(how="any")
    bra.set_index(0, inplace=True)
    bra.index = pd.to_datetime(bra.index.str[:-4],
                               format="%Y-%m-%dT%H:%M:%S").tz_localize(None)
    bra.columns = ["Brasil"]

    output = arg.join(bra, how="left").interpolate(method="linear",
                                                   limit_area="inside")
    output.rename_axis(None, inplace=True)


    metadata._set(output, area="Regional", currency="USD",
                  inf_adj="No", seas_adj="NSA", unit="Tasa",
                  ts_type="-", cumperiods=1)
    metadata._modify_multiindex(output, levels=[3, 5],
                                new_arrays=[["ARS", "ARS", "BRL"],
                                            ["ARS/USD", "ARS/USD", "BRL/USD"]])

    return output


@retry(
    retry_on_exceptions=(HTTPError, URLError),
    max_calls_total=4,
    retry_window_after_first_call_in_seconds=60,
)
def policy_rates() -> pd.DataFrame:
    """Get central bank policy interest rates data.

    Countries/aggregates selected are Argentina and Brazil.

    Returns
    -------
    Daily policy interest rates : pd.DataFrame

    """
    name = "regional_policy_rates"

    r = requests.get(urls[name]["dl"]["main"])
    temp_dir = tempfile.TemporaryDirectory()
    with zipfile.ZipFile(BytesIO(r.content), "r") as f:
        f.extractall(path=temp_dir.name)
        path_temp = path.join(temp_dir.name,
                              "WEBSTATS_CBPOL_D_DATAFLOW_csv_row.csv")
        raw = pd.read_csv(path_temp, usecols=[0, 1, 3], index_col=0,
                          header=2, parse_dates=True).dropna(how="all")
    output = (raw.apply(pd.to_numeric, errors="coerce")
              .interpolate(method="linear", limit_area="inside"))
    output.columns = ["Argentina", "Brasil"]
    output.rename_axis(None, inplace=True)

    metadata._set(output, area="Regional", currency="-",
                  inf_adj="No", seas_adj="NSA", unit="Tasa",
                  ts_type="-", cumperiods=1)
    metadata._modify_multiindex(output, levels=[3],
                                new_arrays=[["ARS", "BRL"]])

    return output


@retry(
    retry_on_exceptions=(HTTPError, URLError),
    max_calls_total=12,
    retry_window_after_first_call_in_seconds=60,
)
def stocks(pipeline: Optional[Pipeline] = None) -> pd.DataFrame:
    """Get stock market index data in USD terms.

    Indexes selected are MERVAL and BOVESPA.

    Parameters
    ----------
    pipeline : econuy.core.Pipeline or None, default None
        An instance of the econuy Pipeline class.

    Returns
    -------
    Daily stock market index in USD terms: pd.DataFrame

    """
    name = "regional_stocks"

    end_date_dt = dt.datetime(2000, 1, 1)
    start_date_dt = dt.datetime(2000, 1, 1)
    aux = []
    while end_date_dt < dt.datetime.now():
        end_date_dt = start_date_dt + dt.timedelta(days=5000)
        params = {
            "curr_id": "13376",
            "smlID": str(randint(1000000, 99999999)),
            "header": "S&amp;P Merval Historical Data",
            "st_date": start_date_dt.strftime("%m/%d/%Y"),
            "end_date": end_date_dt.strftime("%m/%d/%Y"),
            "interval_sec": "Daily",
            "sort_col": "date",
            "sort_ord": "DESC",
            "action": "historical_data"
        }
        r = requests.post(urls[name]["dl"]["arg"],
                          headers=investing_headers, data=params)
        aux.append(pd.read_html(r.content, match="Price",
                                index_col=0, parse_dates=True)[0])
        start_date_dt = end_date_dt + dt.timedelta(days=1)
    arg = pd.concat(aux, axis=0)[["Price"]].sort_index()

    bra = pd.read_csv(urls[name]["dl"]["bra"], index_col=0,
                      parse_dates=True)[["Close"]]
    bra = bra.loc[bra.index >= "2000-01-01"]

    pipeline.get("regional_nxr")
    converters = pipeline.dataset
    converters.columns = converters.columns.get_level_values(0)
    arg = pd.merge_asof(arg, converters[["Argentina - informal"]],
                        left_index=True, right_index=True)
    arg = (arg.iloc[:, 0] / arg.iloc[:, 1]).to_frame()
    arg.columns = ["Argentina"]
    bra = pd.merge_asof(bra, converters[["Brasil"]],
                        left_index=True, right_index=True)
    bra = (bra.iloc[:, 0] / bra.iloc[:, 1]).to_frame()
    bra.columns = ["Brasil"]

    output = arg.join(bra, how="left").interpolate(method="linear",
                                                   limit_area="inside")
    output.rename_axis(None, inplace=True)
    metadata._set(output, area="Regional", currency="USD",
                  inf_adj="No", seas_adj="NSA",
                  ts_type="-", cumperiods=1)
    output = rebase(output, start_date="2019-01-02").dropna(how="all")

    return output


@retry(
    retry_on_exceptions=(HTTPError, URLError),
    max_calls_total=12,
    retry_window_after_first_call_in_seconds=60,
)
def rxr(pipeline: Optional[Pipeline] = None) -> pd.DataFrame:
    """Get real exchange rates vis-á-vis the US dollar for Argentina and Brasil .

    Returns
    -------
    Monthly real exchange rate : pd.DataFrame

    """
    if pipeline is None:
        pipeline = Pipeline()

    proc = _ifs(pipeline=pipeline)

    output = pd.DataFrame()
    output["Argentina"] = (proc["Argentina - oficial"] * proc["US.PCPI_IX"]
                           / proc["ARG CPI"])
    output["Brasil"] = proc["Brasil"] * proc["US.PCPI_IX"] / proc["BRA CPI"]
    output.rename_axis(None, inplace=True)
    metadata._set(output, area="Regional", currency="-",
                  inf_adj="-", seas_adj="NSA",
                  ts_type="-", cumperiods=1)
    metadata._modify_multiindex(output, levels=[3],
                                new_arrays=[["ARS/USD", "BRL/USD"]])
    output = rebase(output, start_date="2019-01-01",
                    end_date="2019-01-31").dropna(how="all")

    return output


@retry(
    retry_on_exceptions=(HTTPError, URLError),
    max_calls_total=4,
    retry_window_after_first_call_in_seconds=30,
)
def _ifs(pipeline: Pipeline = None) -> pd.DataFrame:
    """Get extra CPI and exchange rate data from the IMF IFS.

    Parameters
    ----------
    pipeline : econuy.core.Pipeline or None, default None
        An instance of the econuy Pipeline class.

    Returns
    -------
    IMF data : pd.DataFrame
        CPI and XR for the US, Brazil and Argentina.

    """
    if pipeline is None:
        pipeline = Pipeline()

    url_ = "http://dataservices.imf.org/REST/SDMX_JSON.svc/CompactData/IFS/M."
    url_extra = ".?startPeriod=1970&endPeriod="
    ifs = []
    for country in ["US", "BR", "AR"]:
        for indicator in ["PCPI_IX", "ENDA_XDC_USD_RATE"]:
            base_url = (f"{url_}{country}.{indicator}{url_extra}"
                        f"{dt.datetime.now().year}")
            r_json = requests.get(base_url).json()
            data = r_json["CompactData"]["DataSet"]["Series"]["Obs"]
            try:
                data = pd.DataFrame(data)
                data.set_index("@TIME_PERIOD", drop=True, inplace=True)
            except ValueError:
                data = pd.DataFrame(np.nan,
                                    index=pd.date_range(start="1970-01-01",
                                                        end=dt.datetime.now(),
                                                        freq="M"),
                                    columns=[f"{country}.{indicator}"])
            if "@OBS_STATUS" in data.columns:
                data.drop("@OBS_STATUS", inplace=True, axis=1)
            data.index = (pd.to_datetime(data.index, format="%Y-%m")
                          + MonthEnd(1))
            data.columns = [f"{country}.{indicator}"]
            ifs.append(data)
    ifs = pd.concat(ifs, axis=1, sort=True).apply(pd.to_numeric)

    pipeline.get("regional_nxr")
    xr = pipeline.dataset
    xr = resample(xr, rule="M", operation="mean")
    xr.columns = xr.columns.get_level_values(0)
    pipeline.get("regional_cpi")
    prices = pipeline.dataset
    prices.columns = ["ARG CPI", "BRA CPI"]

    proc = pd.concat([xr, prices, ifs], axis=1)
    proc["Argentina - oficial"] = np.where(
        pd.isna(proc["Argentina - oficial"]),
        proc["AR.ENDA_XDC_USD_RATE"],
        proc["Argentina - oficial"])
    proc["Argentina - informal"] = np.where(
        pd.isna(proc["Argentina - informal"]),
        proc["AR.ENDA_XDC_USD_RATE"],
        proc["Argentina - informal"])
    proc["Brasil"] = np.where(pd.isna(proc["Brasil"]),
                              proc["BR.ENDA_XDC_USD_RATE"],
                              proc["Brasil"])
    proc = proc[["Argentina - oficial", "Argentina - informal",
                 "Brasil", "ARG CPI", "BRA CPI", "US.PCPI_IX"]]

    return proc



def original():
    from econuy.retrieval import (prices, fiscal_accounts, economic_activity,
                                  labor, external_sector, financial_sector, income)

    return {"natacc_ind_con_nsa": {"description": "Cuentas nacionales: Oferta, precios constantes, series armonizadas (mar-05, T)",
                                   "function": economic_activity.natacc_ind_con_nsa},
            "natacc_gas_con_nsa": {"description": "Cuentas nacionales: Demanda, precios constantes, series armonizadas (mar-05, T)",
                                   "function": economic_activity.natacc_gas_con_nsa},
            "natacc_ind_con_idx_sa": {"description": "Cuentas nacionales: Oferta, índice real, series desestacionalizadas (mar-97, T)",
                                      "function": economic_activity.natacc_ind_con_idx_sa},
            "natacc_ind_con_idx_nsa": {"description": "Cuentas nacionales: Oferta, índice real, series armonizadas (mar-97, T)",
                                       "function": economic_activity.natacc_ind_con_idx_nsa},
            "natacc_ind_cur_nsa": {"description": "Cuentas nacionales: Oferta, precios corrientes, series armonizadas (mar-05, T)",
                                   "function": economic_activity.natacc_ind_cur_nsa},
            "natacc_gdp_cur_nsa": {"description": "Cuentas nacionales: PBI, precios corrientes (mar-97, T)",
                                   "function": economic_activity.natacc_gdp_cur_nsa},
            "industrial_production": {"description": "Producción industrial: por división, agrupaciones y clases (ene-02, M)",
                                      "function": economic_activity.industrial_production},
            "electricity": {"description": "Ventas de energía eléctrica por sector (ene-00, M)",
                            "function": economic_activity.electricity},
            "gasoline": {"description": "Ventas de nafta por departamento (ene-04, M)",
                         "function": economic_activity.gasoline},
            "diesel": {"description": "Ventas de gasoil por departamento (ene-04, M)",
                       "function": economic_activity.diesel},
            "cattle": {"description": "Faena de bovinos (2-ene-05, S)",
                       "function": economic_activity.cattle},
            "milk": {"description": "Remisión de leche a planta (ene-02, M)",
                     "function": economic_activity.milk},
            "cement": {"description": "Ventas de cemento (ene-90, M)",
                       "function": economic_activity.cement},
            "cpi": {"description": "Índice de precios al consumidor - IPC (ene-37, M)",
                    "function": prices.cpi},
            "nxr_monthly": {"description": "Tipo de cambio, interbancario, mensual (abr-72, M)",
                            "function": prices.nxr_monthly},
            "nxr_daily": {"description": "Tipo de cambio, interbancario, diario (3-ene-00, D)",
                          "function": prices.nxr_daily},
            "balance_gps": {"description": "Resultado fiscal: Sector público consolidado (ene-99, M)",
                            "function": fiscal_accounts.balance_gps},
            "balance_nfps": {"description": "Resultado fiscal: Sector público no financiero (ene-99, M)",
                             "function": fiscal_accounts.balance_nfps},
            "balance_cg-bps": {"description": "Resultado fiscal: Gobierno central-BPS (ene-99, M)",
                               "function": fiscal_accounts.balance_cg_bps},
            "balance_pe": {"description": "Resultado fiscal de las empresas públicas",
                           "function": fiscal_accounts.balance_pe},
            "balance_ancap": {"description": "Resultado fiscal: ANCAP (ene-01, M)",
                              "function": fiscal_accounts.balance_ancap},
            "balance_ute": {"description": "Resultado fiscal: UTE (ene-99, M)",
                            "function": fiscal_accounts.balance_ute},
            "balance_antel": {"description": "Resultado fiscal: ANTEL (ene-99, M)",
                              "function": fiscal_accounts.balance_antel},
            "balance_ose": {"description": "Resultado fiscal: OSE (ene-99, M)",
                            "function": fiscal_accounts.balance_ose},
            "tax_revenue": {"description": "Recaudación por impuesto (ene-82, M)",
                      "function": fiscal_accounts.tax_revenue},
            "public_debt_gps": {"description": "Deuda del sector público global: por plazo contractual, residual, moneda y residencia (dic-99, T)",
                                "function": fiscal_accounts.public_debt_gps},
            "public_debt_nfps": {"description": "Deuda del sector público no monetario: por plazo contractual, residual, moneda y residencia (dic-99, T)",
                                 "function": fiscal_accounts.public_debt_nfps},
            "public_debt_cb": {"description": "Deuda del Banco Central: por plazo contractual, residual, moneda y residencia (dic-99, T)",
                               "function": fiscal_accounts.public_debt_cb},
            "public_assets": {"description": "Activos del sectro público (dic-99, T)",
                              "function": fiscal_accounts.public_assets},
            "labor_rates": {"description": "Tasas de actividad, empleo y desempleo por sexo (ene-06, M)",
                            "function": labor.labor_rates},
            "hours_worked": {"description": "Horas promedio trabajadas por sector (ene-06, M)",
                             "function": labor.hours},
            "nominal_wages": {"description": "Salarios nominales: total, públicos y privados (ene-68, M)",
                              "function": labor.nominal_wages},
            "real_wages": {"description": "Salarios reales: total, públicos y privados (ene-68, M)",
                           "function": labor.real_wages},
            "trade_x_prod_val": {"description": "Exportaciones por sector, valor (ene-00, M)",
                                 "function": external_sector.trade_x_prod_val},
            "trade_x_prod_vol": {"description": "Exportaciones por sector, volumen (ene-05, M)",
                                 "function": external_sector.trade_x_prod_vol},
            "trade_x_prod_pri": {"description": "Exportaciones por sector, precio (ene-05, M)",
                                 "function": external_sector.trade_x_prod_pri},
            "trade_x_dest_val": {"description": "Exportaciones por destino, valor (ene-00, M)",
                                 "function": external_sector.trade_x_dest_val},
            "trade_x_dest_vol": {"description": "Exportaciones por destino, volumen (ene-05, M)",
                                 "function": external_sector.trade_x_dest_vol},
            "trade_x_dest_pri": {"description": "Exportaciones por destino, precio (ene-05, M)",
                                 "function": external_sector.trade_x_dest_pri},
            "trade_m_sect_val": {"description": "Importaciones por tipo, valor (ene-00, M)",
                                 "function": external_sector.trade_m_sect_val},
            "trade_m_sect_vol": {"description": "Importaciones por tipo, volumen (ene-05, M)",
                                 "function": external_sector.trade_m_sect_vol},
            "trade_m_sect_pri": {"description": "Importaciones por tipo, precio (ene-05, M)",
                                 "function": external_sector.trade_m_sect_pri},
            "trade_m_orig_val": {"description": "Importaciones por origen, valor (ene-00, M)",
                                 "function": external_sector.trade_m_orig_val},
            "trade_m_orig_vol": {"description": "Importaciones por origen, volumen (ene-05, M)",
                                 "function": external_sector.trade_m_orig_vol},
            "trade_m_orig_pri": {"description": "Importaciones por origen, precio (ene-05, M)",
                                 "function": external_sector.trade_m_orig_pri},
            "rxr_official": {"description": "Tipos de cambio reales, BCU (ene-00, M)",
                             "function": external_sector.rxr_official},
            "reserves": {"description": "Reservas internacionales (26-jun-02, D)",
                         "function": external_sector.reserves},
            "reserves_changes": {"description": "Variación en las reservas del BCU por fuente (2-ene-13, D)",
                                 "function": external_sector.reserves_changes},
            "deposits": {"description": "Depósitos en el sistema bancario (dic-98, M)",
                         "function": financial_sector.deposits},
            "credit": {"description": "Créditos del sistema bancario al sector no financiero (dic-98, M)",
                       "function": financial_sector.credit},
            "interest_rates": {"description": "Tasas de interés activas y pasivas (ene-98, M)",
                               "function": financial_sector.interest_rates},
            "call": {"description": "Tasa call a 1 día (02-ene-02, D)",
                     "function": financial_sector.call_rate},
            "sovereign_risk": {"description": "Uruguay Bond Index - riesgo soberano (01-ene-99, D)",
                               "function": financial_sector.sovereign_risk},
            "income_household": {"description": "Ingreso medio de los hogares sin valor ni aguinaldo (ene-06, M)",
                                 "function": income.income_household},
            "income_capita": {"description": "Ingreso medio per cápita sin valor locativo ni aguinaldo (ene-06, M)",
                              "function": income.income_capita},
            "consumer_confidence": {"description": "Índice de Confianza del Consumidor (ago-07, M)",
                                    "function": income.consumer_confidence}}


def custom():
    from econuy.retrieval import (prices, fiscal_accounts, economic_activity,
                                  labor, external_sector, financial_sector, income,
                                  international, regional)

    return {"core_industrial": {"description": "Producción industrial: total, sin refinería y núcleo (ene-02, M [e])",
                                "function": economic_activity.core_industrial},
            "cpi_measures": {"description": "IPC transable, no transable, subyacente y residual (mar-97, M [e])",
                             "function": prices.cpi_measures},
            "balance_summary": {"description": "Resultado fiscal: Todas las agregaciones, inc. aj. FSS (ene-99, M [e])",
                                "function": fiscal_accounts.balance_summary},
            "net_public_debt": {"description": "Deuda neta del sector público global excl. encajes (dic-99, T [e])",
                                "function": fiscal_accounts.net_public_debt},
            "labor_rates_people": {"description": "Actividad, empleo y desempleo: series extendidas de tasas y personas (ene-91, M [e])",
                                   "function": labor.rates_people},
            "trade_balance": {"description": "Balanza comercial por país (ene-00, M [e])",
                              "function": external_sector.trade_balance},
            "terms_of_trade": {"description": "Términos de intercambio (ene-05, M [e])",
                               "function": external_sector.terms_of_trade},
            "commodity_prices": {"description": "Precios de commodities seleccionados (ene-02, M [e])",
                                 "function": external_sector.commodity_prices},
            "commodity_index": {"description": "Índice econuy de precios de commodities (ene-02, M [e])",
                                "function": external_sector.commodity_index},
            "rxr_custom": {"description": "Tipos de cambio reales, cálculos econuy (dic-79, M [e])",
                           "function": external_sector.rxr_custom},
            "bonds": {"description": "Rendimiento de bonos soberanos (02-jun-03, D [e])",
                      "function": financial_sector.bonds},
            "global_gdp": {"description": "PBI real desestacionalizado de EE.UU., Unión Europea, Japón y China (mar-47, T)",
                           "function": international.gdp},
            "global_stocks": {"description": "Índices bursátiles de EE.UU., Unión Europea, Japón y China (30-dic-27, D)",
                              "function": international.stocks},
            "global_policy_rates": {"description": "Tasas de interés de política monetaria de EE.UU., Eurozona, Japón y China (01-ene-46, D)",
                                    "function": international.policy_rates},
            "global_long_rates": {"description": "Tasas de interés de largo plazo (10 años) de EE.UU., Alemania, Francia, España, Italia, Reino Unido, Japón y China (02-ene-62, D)",
                                  "function": international.long_rates},
            "global_nxr": {"description": "Cotización de monedas de EE.UU., Eurozona, Japón y China (04-ene-71, D)",
                           "function": international.nxr},
            "regional_gdp": {"description": "PBI real desestacionalizado de Argentina y Brasil (mar-93, T)",
                             "function": regional.gdp},
            "regional_monthly_gdp": {"description": "PBI real mensual desestacionalizado de Argentina y Brasil (ene-03, M)",
                                     "function": regional.monthly_gdp},
            "regional_cpi": {"description": "Índice de precios al consumo de Argentina y Brasil (ene-70, M)",
                             "function": regional.cpi},
            "regional_embi_spreads": {"description": "EMBI spread (JP Morgan) de Argentina, Brasil y Global (11-dic-98, D)",
                                      "function": regional.embi_spreads},
            "regional_embi_yields": {"description": "EMBI yield (JP Morgan) de Argentina, Brasil y Global (11-dic-98, D)",
                                     "function": regional.embi_yields},
            "regional_nxr": {"description": "Tipo de cambio nominal diario de Argentina y Brasil (9-abr-02, D)",
                             "function": regional.nxr},
            "regional_policy_rates": {"description": "Tasas de interés de política monetaria de Argentina y Brasil (4-jun-86, D)",
                                      "function": regional.policy_rates},
            "regional_stocks": {"description": "Índices bursátiles en USD de Argentina y Brasil (9-abr-02, D)",
                                "function": regional.stocks},
            "regional_rxr": {"description": "Tipo de cambio real vs. EE.UU. (dic-70, M)",
                             "function": regional.rxr},
            "_lin_gdp": {"description": "PBI corriente acumulado mensual linealizado en pesos y USD con proyecciones",
                         "function": economic_activity._lin_gdp}
            }



import os
from pathlib import Path

import pytest

from econuy.session import Session
from econuy.utils import get_project_root, metadata, ops


def trim_rows(input: str, output: str, rows: int = 10):
    for f in os.listdir(input):
        if f.endswith(".csv") and "commodity_weights" not in f:
            df = ops._read(Path(input, f), file_fmt="csv",
                           multiindex="included")
            df = df.reindex(df.index[:-rows])
            ops._save(df, Path(output, f))


@pytest.mark.parametrize("area",
                         ["economic_activity", "prices", "fiscal_accounts",
                          "labor", "external_sector", "financial_sector",
                          "income", "international", "regional"])
def test_retrieval(area):
    """Download every available dataset and compare with a previous version.

    Up to 5% deviations are accepted.
    """
    location = Path(get_project_root().parent, "tests/test-data")
    s = Session(location=location, always_save=False)
    s.get_bulk(names=area)
    for k, v in s.datasets.items():
        print(k)
        test_path = Path(location, f"{k}.csv")
        test = ops._read(test_path)
        metadata._set(test)
        compare = v.reindex(test.index)
        div = test / compare
        final = div[((div > 1.05) | (div < 0.95)).any(1)]
        assert len(final) == 0



from os import path

import numpy as np
import pandas as pd
from pandas.tseries.frequencies import infer_freq
import pytest
from sqlalchemy import create_engine

from econuy import transform
from econuy.session import Session
from econuy.utils import metadata

CUR_DIR = path.abspath(path.dirname(__file__))
TEST_DIR = path.join(CUR_DIR, "test-data")
TEST_CON = create_engine("sqlite://").connect()


def create_dummy_df(freq, periods=200, area="Test", currency="Test",
                    inf_adj="Test", unit="Test", seas_adj="Test",
                    ts_type="Test", cumperiods=1):
    dates = pd.date_range("2000-01-31", periods=periods, freq=freq)
    cols = ["A", "B", "C"]
    data = np.hstack([np.random.uniform(-100, 100, [periods, 1]),
                      np.random.uniform(1, 50, [periods, 1]),
                      np.random.uniform(-100, -50, [periods, 1])])
    output = pd.DataFrame(index=dates, columns=cols, data=data)
    metadata._set(output, area=area, currency=currency,
                  inf_adj=inf_adj, unit=unit, seas_adj=seas_adj,
                  ts_type=ts_type, cumperiods=cumperiods)
    return output


@pytest.mark.parametrize("freq,ts_type,operation,period,pd_per,pd_rol",
                         [("A-DEC", "Flujo", "chg", "last", 1, 1),
                          ("Q-DEC", "Stock", "diff", "inter", 4, 1),
                          ("M", "Flujo", "chg", "annual", 12, 12)])
def test_chg_diff(freq, ts_type, operation, period, pd_per, pd_rol):
    df = create_dummy_df(freq=freq, ts_type=ts_type)
    s = Session(TEST_CON)
    s._datasets.update({"dummy": df})
    s.chg_diff(select="all", operation=operation, period=period)
    compare = s.datasets["dummy"]
    if ts_type == "Flujo" and period == "annual" and pd_rol > 1:
        df = df.rolling(window=pd_rol, min_periods=pd_rol).sum()
    if operation == "chg":
        df = df.pct_change(pd_per).mul(100)
    else:
        df = df.diff(pd_per)
    compare.index, compare.columns = df.index, df.columns
    assert df.equals(compare)


@pytest.mark.parametrize("freq,ts_type,operation,window,pd_rol",
                         [("M", "Flujo", "sum", 3, 3),
                          ("Q-DEC", "Flujo", "mean", None, 4)])
def test_rolling(freq, ts_type, operation, window, pd_rol):
    df = create_dummy_df(freq=freq, ts_type=ts_type)
    s = Session()
    s._datasets["dummy"] = df
    s.rolling(window=window, operation=operation, select="dummy")
    compare = s.datasets["dummy"]
    if operation == "sum":
        df = df.rolling(window=pd_rol, min_periods=pd_rol).sum()
    else:
        df = df.rolling(window=pd_rol, min_periods=pd_rol).mean()
    compare.index, compare.columns = df.index, df.columns
    assert df.equals(compare)


@pytest.mark.parametrize("freq,ts_type,periods,cumperiods,rule,operation",
                         [("M", "Flujo", 192, 1, "A-DEC", "sum"),
                          ("Q-DEC", "Flujo", 102, 2, "A-DEC", "mean"),
                          ("Q-DEC", "Stock", 102, 1, "A-DEC", "last"),
                          ("A-DEC", "Flujo", 102, 1, "M", "upsample")])
def test_resample(freq, ts_type, periods, cumperiods, rule, operation):
    df = create_dummy_df(freq=freq, ts_type=ts_type, periods=periods,
                         cumperiods=cumperiods)
    s = Session()
    s._datasets["dummy"] = df
    s.resample(rule=rule, operation=operation, select=0)
    compare = s._datasets["dummy"]
    if operation == "sum":
        df = df.resample(rule=rule).sum()
    elif operation == "mean":
        df = df.resample(rule=rule).mean()
    elif operation == "last":
        df = df.resample(rule=rule).last()
    else:
        df = df.resample(rule=rule).last().interpolate("linear")

    pd_freqs = {"M": 12, "Q-DEC": 4, "A-DEC": 1}
    if periods % pd_freqs[freq] != 0 and operation != "upsample":
        rows = (periods // pd_freqs[freq])
        df = df.iloc[:rows, :]
    compare.index, compare.columns = df.index, df.columns
    assert df.equals(compare)


@pytest.mark.parametrize("component,method,fallback,trading,outlier",
                         [("trend", "x13", "loess", True, True),
                         ("seas", "x13", "ma", True, False),
                         ("trend", "loess", "ma", False, True),
                         ("trend", "ma", "loess", False, False)])
def test_decompose(component, method, fallback, trading, outlier):
    df = pd.DataFrame(index=pd.date_range("2000-01-01", periods=100,
                                          freq="Q-DEC"),
                      data=np.random.exponential(2, 100).cumsum(),
                      columns=["Exponential"])
    df["Real"] = df["Exponential"]
    df.loc[df.index.month == 12,
           "Real"] = (df.loc[df.index.month == 12, "Real"].
                      multiply(np.random.uniform(1.06, 1.14)))
    df.loc[df.index.month == 6,
           "Real"] = (df.loc[df.index.month == 6, "Real"].
                      multiply(np.random.uniform(0.94, 0.96)))
    df.loc[df.index.month == 3,
           "Real"] = (df.loc[df.index.month == 3, "Real"].
                      multiply(np.random.uniform(1.04, 1.06)))
    df.drop("Exponential", axis=1, inplace=True)
    df = df.add(np.random.normal(0, 1, 100), axis=0)
    metadata._set(df, seas_adj="NSA")
    s = Session()
    s._datasets["test"] = df
    s.decompose(component=component, trading=trading, method=method,
                outlier=outlier, fallback=fallback,
                ignore_warnings=True)
    assert s._datasets["test"].std().values <= df.std().values


@pytest.mark.parametrize("freq,start_date,end_date,base",
                         [("M", "2004-01-01", None, 100),
                          ("Q-DEC", "2004-01-01", "2005-01-01", 100.1)])
def test_rebase(freq, start_date, end_date, base):
    df = create_dummy_df(freq=freq)
    s = Session()
    s._datasets["dummy"] = df
    s.rebase(start_date=start_date, end_date=end_date, base=base, select=[0])
    compare = s._datasets["dummy"]
    if end_date is None:
        start = df.iloc[df.index.get_loc(start_date, method="nearest")].name
        df = df / df.loc[start] * base
    else:
        df = df / df.loc[start_date:end_date].mean() * base
    compare.index, compare.columns = df.index, df.columns
    assert df.equals(compare)


@pytest.mark.parametrize("freq,ts_type,periods,cumperiods",
                         [("D", "Flujo", 600, 1),
                          ("M", "Stock", 48, 1),
                          ("W-SUN", "Stock", 200, 1),
                          ("Q-DEC", "Flujo", 24, 4)])
def test_convert_usd(freq, ts_type, periods, cumperiods):
    df = create_dummy_df(freq=freq, periods=periods, ts_type=ts_type,
                         currency="UYU", cumperiods=cumperiods)
    s = Session(location=TEST_DIR, download=False)
    s._datasets["dummy"] = df
    s.convert(flavor="usd")
    s.get("nxr_monthly")
    compare = s.datasets["dummy"]
    nxr = s.datasets["nxr_monthly"]
    proc_freq = freq

    if freq in ["D", "W", "B", "W-SUN"]:
        if ts_type == "Stock":
            df = df.resample("M").last()
        else:
            df = df.resample("M").sum()
        proc_freq = "M"

    if ts_type == "Stock":
        nxr = nxr.resample(proc_freq).last().iloc[:, [1]]
    else:
        nxr = nxr.resample(proc_freq).mean().iloc[:, [0]]
        nxr = nxr.rolling(window=cumperiods, min_periods=cumperiods).mean()

    nxr = nxr.reindex(df.index).iloc[:, 0]
    df = df.div(nxr, axis=0)
    compare.index, compare.columns = df.index, df.columns
    assert df.equals(compare)


@pytest.mark.parametrize("freq,periods,ts_type,start_date,end_date",
                         [("M", 100, "Flujo", None, None),
                          ("D", 1200, "Flujo", "2002-01-01", None),
                          ("Q-DEC", 40, "Stock", "2002-01-01", "2002-12-31")])
def test_convert_real(freq, periods, ts_type, start_date, end_date):
    df = create_dummy_df(freq=freq, periods=periods, ts_type=ts_type,
                         currency="UYU")
    s = Session(location=TEST_DIR, download=False)
    s._datasets["dummy"] = df
    s.convert(flavor="real", start_date=start_date, end_date=end_date)
    s.get("cpi")
    compare = s.datasets["dummy"]
    cpi = s.datasets["cpi"]
    proc_freq = freq

    if freq in ["D", "W", "B", "W-SUN"]:
        if ts_type == "Stock":
            df = df.resample("M").last()
        else:
            df = df.resample("M").sum()
        proc_freq = "M"
    save_index = df.index
    cpi = cpi.resample(rule=proc_freq).mean().iloc[:, 0]

    if start_date is None:
        df = df.div(cpi, axis=0)
    elif end_date is None:
        month = df.iloc[df.index.get_loc(start_date, method="nearest")].name
        df = df.div(cpi, axis=0) * cpi.loc[month]
    else:
        df = df.div(cpi, axis=0) * cpi[start_date:end_date].mean()
    df = df.reindex(save_index)
    compare.index, compare.columns = df.index, df.columns
    assert df.equals(compare)


@pytest.mark.parametrize("freq,ts_type,periods,cumperiods,currency",
                         [("D", "Flujo", 600, 1, "UYU"),
                          ("M", "Stock", 48, 1, "USD"),
                          ("A-DEC", "Stock", 10, 1, "USD"),
                          ("Q-DEC", "Flujo", 24, 4, "UYU")])
def test_convert_gdp(freq, ts_type, periods, cumperiods, currency):
    df = create_dummy_df(freq=freq, periods=periods, ts_type=ts_type,
                         currency=currency, cumperiods=cumperiods)
    s = Session(location=TEST_DIR, download=False)
    s._datasets["dummy"] = df
    s.convert(flavor="gdp")
    s.get("_lin_gdp")
    compare = s.datasets["dummy"]
    gdp = s.datasets["_lin_gdp"]

    if freq in ["M", "MS"]:
        gdp = gdp.resample(rule=freq).interpolate("linear")
        if cumperiods != 12 and ts_type == "Flujo":
            converter = int(12 / cumperiods)
            df = df.rolling(window=converter).sum()
    elif freq in ["Q", "Q-DEC"]:
        gdp = gdp.resample(freq, convention="end").asfreq()
        if cumperiods != 4 and ts_type == "Flujo":
            converter = int(4 / cumperiods)
            df = df.rolling(window=converter).sum()
    elif freq in ["A", "A-DEC"]:
        gdp = gdp.resample(freq, convention="end").asfreq()
    else:
        if ts_type == "Flujo":
            df = df.resample("M").sum()
        else:
            df = df.resample("M").mean()
        gdp = gdp.resample(rule="M").interpolate("linear")

    if currency == "USD":
        gdp = gdp.iloc[:, 1].to_frame()
    else:
        gdp = gdp.iloc[:, 0].to_frame()

    gdp = gdp.reindex(df.index).iloc[:, 0]
    df = df.div(gdp, axis=0).multiply(100)
    compare.index, compare.columns = df.index, df.columns
    div = df / compare
    final = div[((div > 1.01) | (div < 0.99)).any(1)]
    assert len(final) == 0
